from io import BytesIO

from PIL import Image

from .helper import PillowTestCase

PIL151 = b"""
#define basic_width 32
#define basic_height 32
static char basic_bits[] = {
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00,
0x80, 0xff, 0xff, 0x01, 0x40, 0x00, 0x00, 0x02,
0x20, 0x00, 0x00, 0x04, 0x20, 0x00, 0x00, 0x04, 0x10, 0x00, 0x00, 0x08,
0x10, 0x00, 0x00, 0x08,
0x10, 0x00, 0x00, 0x08, 0x10, 0x00, 0x00, 0x08,
0x10, 0x00, 0x00, 0x08, 0x10, 0x00, 0x00, 0x08, 0x10, 0x00, 0x00, 0x08,
0x20, 0x00, 0x00, 0x04,
0x20, 0x00, 0x00, 0x04, 0x40, 0x00, 0x00, 0x02,
0x80, 0xff, 0xff, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00,
};
"""


class TestFileXbm(PillowTestCase):
    def test_pil151(self):
        with Image.open(BytesIO(PIL151)) as im:
            im.load()
            self.assertEqual(im.mode, "1")
            self.assertEqual(im.size, (32, 32))

    def test_open(self):
        # Arrange
        # Created with `convert hopper.png hopper.xbm`
        filename = "Tests/images/hopper.xbm"

        # Act
        with Image.open(filename) as im:

            # Assert
            self.assertEqual(im.mode, "1")
            self.assertEqual(im.size, (128, 128))

    def test_open_filename_with_underscore(self):
        # Arrange
        # Created with `convert hopper.png hopper_underscore.xbm`
        filename = "Tests/images/hopper_underscore.xbm"

        # Act
        with Image.open(filename) as im:

            # Assert
            self.assertEqual(im.mode, "1")
            self.assertEqual(im.size, (128, 128))



"""
Helper functions.
"""

import logging
import os
import subprocess
import sys
import tempfile
import unittest
from io import BytesIO

from PIL import Image, ImageMath

logger = logging.getLogger(__name__)


HAS_UPLOADER = False

if os.environ.get("SHOW_ERRORS", None):
    # local img.show for errors.
    HAS_UPLOADER = True

    class test_image_results:
        @staticmethod
        def upload(a, b):
            a.show()
            b.show()


elif "GITHUB_ACTIONS" in os.environ:
    HAS_UPLOADER = True

    class test_image_results:
        @staticmethod
        def upload(a, b):
            dir_errors = os.path.join(os.path.dirname(__file__), "errors")
            os.makedirs(dir_errors, exist_ok=True)
            tmpdir = tempfile.mkdtemp(dir=dir_errors)
            a.save(os.path.join(tmpdir, "a.png"))
            b.save(os.path.join(tmpdir, "b.png"))
            return tmpdir


else:
    try:
        import test_image_results

        HAS_UPLOADER = True
    except ImportError:
        pass


def convert_to_comparable(a, b):
    new_a, new_b = a, b
    if a.mode == "P":
        new_a = Image.new("L", a.size)
        new_b = Image.new("L", b.size)
        new_a.putdata(a.getdata())
        new_b.putdata(b.getdata())
    elif a.mode == "I;16":
        new_a = a.convert("I")
        new_b = b.convert("I")
    return new_a, new_b


class PillowTestCase(unittest.TestCase):
    def delete_tempfile(self, path):
        try:
            os.remove(path)
        except OSError:
            pass  # report?

    def assert_deep_equal(self, a, b, msg=None):
        try:
            self.assertEqual(
                len(a),
                len(b),
                msg or "got length {}, expected {}".format(len(a), len(b)),
            )
            self.assertTrue(
                all(x == y for x, y in zip(a, b)),
                msg or "got {}, expected {}".format(a, b),
            )
        except Exception:
            self.assertEqual(a, b, msg)

    def assert_image(self, im, mode, size, msg=None):
        if mode is not None:
            self.assertEqual(
                im.mode,
                mode,
                msg or "got mode {!r}, expected {!r}".format(im.mode, mode),
            )

        if size is not None:
            self.assertEqual(
                im.size,
                size,
                msg or "got size {!r}, expected {!r}".format(im.size, size),
            )

    def assert_image_equal(self, a, b, msg=None):
        self.assertEqual(
            a.mode, b.mode, msg or "got mode {!r}, expected {!r}".format(a.mode, b.mode)
        )
        self.assertEqual(
            a.size, b.size, msg or "got size {!r}, expected {!r}".format(a.size, b.size)
        )
        if a.tobytes() != b.tobytes():
            if HAS_UPLOADER:
                try:
                    url = test_image_results.upload(a, b)
                    logger.error("Url for test images: %s" % url)
                except Exception:
                    pass

            self.fail(msg or "got different content")

    def assert_image_equal_tofile(self, a, filename, msg=None, mode=None):
        with Image.open(filename) as img:
            if mode:
                img = img.convert(mode)
            self.assert_image_equal(a, img, msg)

    def assert_image_similar(self, a, b, epsilon, msg=None):
        epsilon = float(epsilon)
        self.assertEqual(
            a.mode, b.mode, msg or "got mode {!r}, expected {!r}".format(a.mode, b.mode)
        )
        self.assertEqual(
            a.size, b.size, msg or "got size {!r}, expected {!r}".format(a.size, b.size)
        )

        a, b = convert_to_comparable(a, b)

        diff = 0
        for ach, bch in zip(a.split(), b.split()):
            chdiff = ImageMath.eval("abs(a - b)", a=ach, b=bch).convert("L")
            diff += sum(i * num for i, num in enumerate(chdiff.histogram()))

        ave_diff = float(diff) / (a.size[0] * a.size[1])
        try:
            self.assertGreaterEqual(
                epsilon,
                ave_diff,
                (msg or "")
                + " average pixel value difference %.4f > epsilon %.4f"
                % (ave_diff, epsilon),
            )
        except Exception as e:
            if HAS_UPLOADER:
                try:
                    url = test_image_results.upload(a, b)
                    logger.error("Url for test images: %s" % url)
                except Exception:
                    pass
            raise e

    def assert_image_similar_tofile(self, a, filename, epsilon, msg=None, mode=None):
        with Image.open(filename) as img:
            if mode:
                img = img.convert(mode)
            self.assert_image_similar(a, img, epsilon, msg)

    def assert_warning(self, warn_class, func, *args, **kwargs):
        import warnings

        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")

            # Hopefully trigger a warning.
            result = func(*args, **kwargs)

            # Verify some things.
            if warn_class is None:
                self.assertEqual(
                    len(w), 0, "Expected no warnings, got %s" % [v.category for v in w]
                )
            else:
                self.assertGreaterEqual(len(w), 1)
                found = False
                for v in w:
                    if issubclass(v.category, warn_class):
                        found = True
                        break
                self.assertTrue(found)
        return result

    def assert_all_same(self, items, msg=None):
        self.assertEqual(items.count(items[0]), len(items), msg)

    def assert_not_all_same(self, items, msg=None):
        self.assertNotEqual(items.count(items[0]), len(items), msg)

    def assert_tuple_approx_equal(self, actuals, targets, threshold, msg):
        """Tests if actuals has values within threshold from targets"""

        value = True
        for i, target in enumerate(targets):
            value *= target - threshold <= actuals[i] <= target + threshold

        self.assertTrue(value, msg + ": " + repr(actuals) + " != " + repr(targets))

    def skipKnownBadTest(self, msg=None):
        # Skip if PILLOW_RUN_KNOWN_BAD is not true in the environment.
        if os.environ.get("PILLOW_RUN_KNOWN_BAD", False):
            print(os.environ.get("PILLOW_RUN_KNOWN_BAD", False))
            return

        self.skipTest(msg or "Known Bad Test")

    def tempfile(self, template):
        assert template[:5] in ("temp.", "temp_")
        fd, path = tempfile.mkstemp(template[4:], template[:4])
        os.close(fd)

        self.addCleanup(self.delete_tempfile, path)
        return path

    def open_withImagemagick(self, f):
        if not imagemagick_available():
            raise OSError()

        outfile = self.tempfile("temp.png")
        if command_succeeds([IMCONVERT, f, outfile]):
            return Image.open(outfile)
        raise OSError()


@unittest.skipIf(sys.platform.startswith("win32"), "requires Unix or macOS")
class PillowLeakTestCase(PillowTestCase):
    # requires unix/macOS
    iterations = 100  # count
    mem_limit = 512  # k

    def _get_mem_usage(self):
        """
        Gets the RUSAGE memory usage, returns in K. Encapsulates the difference
        between macOS and Linux rss reporting

        :returns: memory usage in kilobytes
        """

        from resource import getrusage, RUSAGE_SELF

        mem = getrusage(RUSAGE_SELF).ru_maxrss
        if sys.platform == "darwin":
            # man 2 getrusage:
            #     ru_maxrss
            # This is the maximum resident set size utilized (in bytes).
            return mem / 1024  # Kb
        else:
            # linux
            # man 2 getrusage
            #        ru_maxrss (since Linux 2.6.32)
            #  This is the maximum resident set size used (in kilobytes).
            return mem  # Kb

    def _test_leak(self, core):
        start_mem = self._get_mem_usage()
        for cycle in range(self.iterations):
            core()
            mem = self._get_mem_usage() - start_mem
            msg = "memory usage limit exceeded in iteration %d" % cycle
            self.assertLess(mem, self.mem_limit, msg)


# helpers


def fromstring(data):
    return Image.open(BytesIO(data))


def tostring(im, string_format, **options):
    out = BytesIO()
    im.save(out, string_format, **options)
    return out.getvalue()


def hopper(mode=None, cache={}):
    if mode is None:
        # Always return fresh not-yet-loaded version of image.
        # Operations on not-yet-loaded images is separate class of errors
        # what we should catch.
        return Image.open("Tests/images/hopper.ppm")
    # Use caching to reduce reading from disk but so an original copy is
    # returned each time and the cached image isn't modified by tests
    # (for fast, isolated, repeatable tests).
    im = cache.get(mode)
    if im is None:
        if mode == "F":
            im = hopper("L").convert(mode)
        elif mode[:4] == "I;16":
            im = hopper("I").convert(mode)
        else:
            im = hopper().convert(mode)
        cache[mode] = im
    return im.copy()


def command_succeeds(cmd):
    """
    Runs the command, which must be a list of strings. Returns True if the
    command succeeds, or False if an OSError was raised by subprocess.Popen.
    """
    try:
        subprocess.call(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
    except OSError:
        return False
    return True


def djpeg_available():
    return command_succeeds(["djpeg", "-version"])


def cjpeg_available():
    return command_succeeds(["cjpeg", "-version"])


def netpbm_available():
    return command_succeeds(["ppmquant", "--version"]) and command_succeeds(
        ["ppmtogif", "--version"]
    )


def imagemagick_available():
    return IMCONVERT and command_succeeds([IMCONVERT, "-version"])


def on_appveyor():
    return "APPVEYOR" in os.environ


def on_github_actions():
    return "GITHUB_ACTIONS" in os.environ


def on_ci():
    # Travis and AppVeyor have "CI"
    # Azure Pipelines has "TF_BUILD"
    # GitHub Actions has "GITHUB_ACTIONS"
    return (
        "CI" in os.environ or "TF_BUILD" in os.environ or "GITHUB_ACTIONS" in os.environ
    )


def is_win32():
    return sys.platform.startswith("win32")


def is_pypy():
    return hasattr(sys, "pypy_translation_info")


if sys.platform == "win32":
    IMCONVERT = os.environ.get("MAGICK_HOME", "")
    if IMCONVERT:
        IMCONVERT = os.path.join(IMCONVERT, "convert.exe")
else:
    IMCONVERT = "convert"


def distro():
    if os.path.exists("/etc/os-release"):
        with open("/etc/os-release", "r") as f:
            for line in f:
                if "ID=" in line:
                    return line.strip().split("=")[1]


class cached_property:
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls=None):
        result = instance.__dict__[self.func.__name__] = self.func(instance)
        return result



import unittest

from PIL import Image, TarIO

from .helper import PillowTestCase, is_pypy

codecs = dir(Image.core)

# Sample tar archive
TEST_TAR_FILE = "Tests/images/hopper.tar"


class TestFileTar(PillowTestCase):
    def setUp(self):
        if "zip_decoder" not in codecs and "jpeg_decoder" not in codecs:
            self.skipTest("neither jpeg nor zip support available")

    def test_sanity(self):
        for codec, test_path, format in [
            ["zip_decoder", "hopper.png", "PNG"],
            ["jpeg_decoder", "hopper.jpg", "JPEG"],
        ]:
            if codec in codecs:
                with TarIO.TarIO(TEST_TAR_FILE, test_path) as tar:
                    with Image.open(tar) as im:
                        im.load()
                        self.assertEqual(im.mode, "RGB")
                        self.assertEqual(im.size, (128, 128))
                        self.assertEqual(im.format, format)

    @unittest.skipIf(is_pypy(), "Requires CPython")
    def test_unclosed_file(self):
        def open():
            TarIO.TarIO(TEST_TAR_FILE, "hopper.jpg")

        self.assert_warning(ResourceWarning, open)

    def test_close(self):
        def open():
            tar = TarIO.TarIO(TEST_TAR_FILE, "hopper.jpg")
            tar.close()

        self.assert_warning(None, open)

    def test_contextmanager(self):
        def open():
            with TarIO.TarIO(TEST_TAR_FILE, "hopper.jpg"):
                pass

        self.assert_warning(None, open)



import sys
import unittest

from PIL import Image

from .helper import PillowTestCase, is_win32

try:
    import numpy
except ImportError:
    numpy = None


@unittest.skipIf(is_win32(), "Win32 does not call map_buffer")
class TestMap(PillowTestCase):
    def test_overflow(self):
        # There is the potential to overflow comparisons in map.c
        # if there are > SIZE_MAX bytes in the image or if
        # the file encodes an offset that makes
        # (offset + size(bytes)) > SIZE_MAX

        # Note that this image triggers the decompression bomb warning:
        max_pixels = Image.MAX_IMAGE_PIXELS
        Image.MAX_IMAGE_PIXELS = None

        # This image hits the offset test.
        with Image.open("Tests/images/l2rgb_read.bmp") as im:
            with self.assertRaises((ValueError, MemoryError, IOError)):
                im.load()

        Image.MAX_IMAGE_PIXELS = max_pixels

    @unittest.skipIf(sys.maxsize <= 2 ** 32, "requires 64-bit system")
    @unittest.skipIf(numpy is None, "Numpy is not installed")
    def test_ysize(self):
        # Should not raise 'Integer overflow in ysize'
        arr = numpy.zeros((46341, 46341), dtype=numpy.uint8)
        Image.fromarray(arr)



from .helper import PillowTestCase, hopper


class TestImageEntropy(PillowTestCase):
    def test_entropy(self):
        def entropy(mode):
            return hopper(mode).entropy()

        self.assertAlmostEqual(entropy("1"), 0.9138803254693582)
        self.assertAlmostEqual(entropy("L"), 7.063008716585465)
        self.assertAlmostEqual(entropy("I"), 7.063008716585465)
        self.assertAlmostEqual(entropy("F"), 7.063008716585465)
        self.assertAlmostEqual(entropy("P"), 5.0530452472519745)
        self.assertAlmostEqual(entropy("RGB"), 8.821286587714319)
        self.assertAlmostEqual(entropy("RGBA"), 7.42724306524488)
        self.assertAlmostEqual(entropy("CMYK"), 7.4272430652448795)
        self.assertAlmostEqual(entropy("YCbCr"), 7.698360534903628)



from PIL import Image, McIdasImagePlugin

from .helper import PillowTestCase


class TestFileMcIdas(PillowTestCase):
    def test_invalid_file(self):
        invalid_file = "Tests/images/flower.jpg"

        self.assertRaises(SyntaxError, McIdasImagePlugin.McIdasImageFile, invalid_file)

    def test_valid_file(self):
        # Arrange
        # https://ghrc.nsstc.nasa.gov/hydro/details/cmx3g8
        # https://ghrc.nsstc.nasa.gov/pub/fieldCampaigns/camex3/cmx3g8/browse/
        test_file = "Tests/images/cmx3g8_wv_1998.260_0745_mcidas.ara"
        saved_file = "Tests/images/cmx3g8_wv_1998.260_0745_mcidas.png"

        # Act
        with Image.open(test_file) as im:
            im.load()

            # Assert
            self.assertEqual(im.format, "MCIDAS")
            self.assertEqual(im.mode, "I")
            self.assertEqual(im.size, (1800, 400))
            with Image.open(saved_file) as im2:
                self.assert_image_equal(im, im2)



from PIL import Image

from .helper import PillowTestCase, hopper


class TestImageFromBytes(PillowTestCase):
    def test_sanity(self):
        im1 = hopper()
        im2 = Image.frombytes(im1.mode, im1.size, im1.tobytes())

        self.assert_image_equal(im1, im2)

    def test_not_implemented(self):
        self.assertRaises(NotImplementedError, Image.fromstring)



import os
import unittest

from PIL import Image, SunImagePlugin

from .helper import PillowTestCase, hopper

EXTRA_DIR = "Tests/images/sunraster"


class TestFileSun(PillowTestCase):
    def test_sanity(self):
        # Arrange
        # Created with ImageMagick: convert hopper.jpg hopper.ras
        test_file = "Tests/images/hopper.ras"

        # Act
        with Image.open(test_file) as im:

            # Assert
            self.assertEqual(im.size, (128, 128))

            self.assert_image_similar(im, hopper(), 5)  # visually verified

        invalid_file = "Tests/images/flower.jpg"
        self.assertRaises(SyntaxError, SunImagePlugin.SunImageFile, invalid_file)

    def test_im1(self):
        with Image.open("Tests/images/sunraster.im1") as im:
            with Image.open("Tests/images/sunraster.im1.png") as target:
                self.assert_image_equal(im, target)

    @unittest.skipUnless(os.path.exists(EXTRA_DIR), "Extra image files not installed")
    def test_others(self):
        files = (
            os.path.join(EXTRA_DIR, f)
            for f in os.listdir(EXTRA_DIR)
            if os.path.splitext(f)[1] in (".sun", ".SUN", ".ras")
        )
        for path in files:
            with Image.open(path) as im:
                im.load()
                self.assertIsInstance(im, SunImagePlugin.SunImageFile)
                target_path = "%s.png" % os.path.splitext(path)[0]
                # im.save(target_file)
                with Image.open(target_path) as target:
                    self.assert_image_equal(im, target)



import sys
import unittest

from PIL import Image

from .helper import PillowTestCase

# This test is not run automatically.
#
# It requires > 2gb memory for the >2 gigapixel image generated in the
# second test.  Running this automatically would amount to a denial of
# service on our testing infrastructure.  I expect this test to fail
# on any 32-bit machine, as well as any smallish things (like
# Raspberry Pis).


try:
    import numpy as np
except ImportError:
    raise unittest.SkipTest("numpy not installed")

YDIM = 32769
XDIM = 48000


@unittest.skipIf(sys.maxsize <= 2 ** 32, "requires 64-bit system")
class LargeMemoryNumpyTest(PillowTestCase):
    def _write_png(self, xdim, ydim):
        dtype = np.uint8
        a = np.zeros((xdim, ydim), dtype=dtype)
        f = self.tempfile("temp.png")
        im = Image.fromarray(a, "L")
        im.save(f)

    def test_large(self):
        """ succeeded prepatch"""
        self._write_png(XDIM, YDIM)

    def test_2gpx(self):
        """failed prepatch"""
        self._write_png(XDIM, XDIM)


if __name__ == "__main__":
    unittest.main()



import unittest
import zlib
from io import BytesIO

from PIL import Image, ImageFile, PngImagePlugin

from .helper import PillowTestCase

TEST_FILE = "Tests/images/png_decompression_dos.png"


class TestPngDos(PillowTestCase):
    def test_ignore_dos_text(self):
        ImageFile.LOAD_TRUNCATED_IMAGES = True

        try:
            im = Image.open(TEST_FILE)
            im.load()
        finally:
            ImageFile.LOAD_TRUNCATED_IMAGES = False

        for s in im.text.values():
            self.assertLess(len(s), 1024 * 1024, "Text chunk larger than 1M")

        for s in im.info.values():
            self.assertLess(len(s), 1024 * 1024, "Text chunk larger than 1M")

    def test_dos_text(self):

        try:
            im = Image.open(TEST_FILE)
            im.load()
        except ValueError as msg:
            self.assertTrue(msg, "Decompressed Data Too Large")
            return

        for s in im.text.values():
            self.assertLess(len(s), 1024 * 1024, "Text chunk larger than 1M")

    def test_dos_total_memory(self):
        im = Image.new("L", (1, 1))
        compressed_data = zlib.compress(b"a" * 1024 * 1023)

        info = PngImagePlugin.PngInfo()

        for x in range(64):
            info.add_text("t%s" % x, compressed_data, zip=True)
            info.add_itxt("i%s" % x, compressed_data, zip=True)

        b = BytesIO()
        im.save(b, "PNG", pnginfo=info)
        b.seek(0)

        try:
            im2 = Image.open(b)
        except ValueError as msg:
            self.assertIn("Too much memory", msg)
            return

        total_len = 0
        for txt in im2.text.values():
            total_len += len(txt)
        self.assertLess(
            total_len, 64 * 1024 * 1024, "Total text chunks greater than 64M"
        )


if __name__ == "__main__":
    unittest.main()



import unittest
from io import BytesIO

from PIL import Image

from .helper import PillowTestCase, is_pypy

test_files = ["Tests/images/sugarshack.mpo", "Tests/images/frozenpond.mpo"]


class TestFileMpo(PillowTestCase):
    def setUp(self):
        codecs = dir(Image.core)
        if "jpeg_encoder" not in codecs or "jpeg_decoder" not in codecs:
            self.skipTest("jpeg support not available")

    def frame_roundtrip(self, im, **options):
        # Note that for now, there is no MPO saving functionality
        out = BytesIO()
        im.save(out, "MPO", **options)
        test_bytes = out.tell()
        out.seek(0)
        im = Image.open(out)
        im.bytes = test_bytes  # for testing only
        return im

    def test_sanity(self):
        for test_file in test_files:
            with Image.open(test_file) as im:
                im.load()
                self.assertEqual(im.mode, "RGB")
                self.assertEqual(im.size, (640, 480))
                self.assertEqual(im.format, "MPO")

    @unittest.skipIf(is_pypy(), "Requires CPython")
    def test_unclosed_file(self):
        def open():
            im = Image.open(test_files[0])
            im.load()

        self.assert_warning(ResourceWarning, open)

    def test_closed_file(self):
        def open():
            im = Image.open(test_files[0])
            im.load()
            im.close()

        self.assert_warning(None, open)

    def test_context_manager(self):
        def open():
            with Image.open(test_files[0]) as im:
                im.load()

        self.assert_warning(None, open)

    def test_app(self):
        for test_file in test_files:
            # Test APP/COM reader (@PIL135)
            with Image.open(test_file) as im:
                self.assertEqual(im.applist[0][0], "APP1")
                self.assertEqual(im.applist[1][0], "APP2")
                self.assertEqual(
                    im.applist[1][1][:16],
                    b"MPF\x00MM\x00*\x00\x00\x00\x08\x00\x03\xb0\x00",
                )
                self.assertEqual(len(im.applist), 2)

    def test_exif(self):
        for test_file in test_files:
            with Image.open(test_file) as im:
                info = im._getexif()
                self.assertEqual(info[272], "Nintendo 3DS")
                self.assertEqual(info[296], 2)
                self.assertEqual(info[34665], 188)

    def test_frame_size(self):
        # This image has been hexedited to contain a different size
        # in the EXIF data of the second frame
        with Image.open("Tests/images/sugarshack_frame_size.mpo") as im:
            self.assertEqual(im.size, (640, 480))

            im.seek(1)
            self.assertEqual(im.size, (680, 480))

    def test_parallax(self):
        # Nintendo
        with Image.open("Tests/images/sugarshack.mpo") as im:
            exif = im.getexif()
            self.assertEqual(
                exif.get_ifd(0x927C)[0x1101]["Parallax"], -44.798187255859375
            )

        # Fujifilm
        with Image.open("Tests/images/fujifilm.mpo") as im:
            im.seek(1)
            exif = im.getexif()
            self.assertEqual(exif.get_ifd(0x927C)[0xB211], -3.125)

    def test_mp(self):
        for test_file in test_files:
            with Image.open(test_file) as im:
                mpinfo = im._getmp()
                self.assertEqual(mpinfo[45056], b"0100")
                self.assertEqual(mpinfo[45057], 2)

    def test_mp_offset(self):
        # This image has been manually hexedited to have an IFD offset of 10
        # in APP2 data, in contrast to normal 8
        with Image.open("Tests/images/sugarshack_ifd_offset.mpo") as im:
            mpinfo = im._getmp()
            self.assertEqual(mpinfo[45056], b"0100")
            self.assertEqual(mpinfo[45057], 2)

    def test_mp_no_data(self):
        # This image has been manually hexedited to have the second frame
        # beyond the end of the file
        with Image.open("Tests/images/sugarshack_no_data.mpo") as im:
            with self.assertRaises(ValueError):
                im.seek(1)

    def test_mp_attribute(self):
        for test_file in test_files:
            with Image.open(test_file) as im:
                mpinfo = im._getmp()
            frameNumber = 0
            for mpentry in mpinfo[45058]:
                mpattr = mpentry["Attribute"]
                if frameNumber:
                    self.assertFalse(mpattr["RepresentativeImageFlag"])
                else:
                    self.assertTrue(mpattr["RepresentativeImageFlag"])
                self.assertFalse(mpattr["DependentParentImageFlag"])
                self.assertFalse(mpattr["DependentChildImageFlag"])
                self.assertEqual(mpattr["ImageDataFormat"], "JPEG")
                self.assertEqual(mpattr["MPType"], "Multi-Frame Image: (Disparity)")
                self.assertEqual(mpattr["Reserved"], 0)
                frameNumber += 1

    def test_seek(self):
        for test_file in test_files:
            with Image.open(test_file) as im:
                self.assertEqual(im.tell(), 0)
                # prior to first image raises an error, both blatant and borderline
                self.assertRaises(EOFError, im.seek, -1)
                self.assertRaises(EOFError, im.seek, -523)
                # after the final image raises an error,
                # both blatant and borderline
                self.assertRaises(EOFError, im.seek, 2)
                self.assertRaises(EOFError, im.seek, 523)
                # bad calls shouldn't change the frame
                self.assertEqual(im.tell(), 0)
                # this one will work
                im.seek(1)
                self.assertEqual(im.tell(), 1)
                # and this one, too
                im.seek(0)
                self.assertEqual(im.tell(), 0)

    def test_n_frames(self):
        with Image.open("Tests/images/sugarshack.mpo") as im:
            self.assertEqual(im.n_frames, 2)
            self.assertTrue(im.is_animated)

    def test_eoferror(self):
        with Image.open("Tests/images/sugarshack.mpo") as im:
            n_frames = im.n_frames

            # Test seeking past the last frame
            self.assertRaises(EOFError, im.seek, n_frames)
            self.assertLess(im.tell(), n_frames)

            # Test that seeking to the last frame does not raise an error
            im.seek(n_frames - 1)

    def test_image_grab(self):
        for test_file in test_files:
            with Image.open(test_file) as im:
                self.assertEqual(im.tell(), 0)
                im0 = im.tobytes()
                im.seek(1)
                self.assertEqual(im.tell(), 1)
                im1 = im.tobytes()
                im.seek(0)
                self.assertEqual(im.tell(), 0)
                im02 = im.tobytes()
                self.assertEqual(im0, im02)
                self.assertNotEqual(im0, im1)

    def test_save(self):
        # Note that only individual frames can be saved at present
        for test_file in test_files:
            with Image.open(test_file) as im:
                self.assertEqual(im.tell(), 0)
                jpg0 = self.frame_roundtrip(im)
                self.assert_image_similar(im, jpg0, 30)
                im.seek(1)
                self.assertEqual(im.tell(), 1)
                jpg1 = self.frame_roundtrip(im)
                self.assert_image_similar(im, jpg1, 30)



from .helper import PillowTestCase, hopper


class TestImageHistogram(PillowTestCase):
    def test_histogram(self):
        def histogram(mode):
            h = hopper(mode).histogram()
            return len(h), min(h), max(h)

        self.assertEqual(histogram("1"), (256, 0, 10994))
        self.assertEqual(histogram("L"), (256, 0, 662))
        self.assertEqual(histogram("I"), (256, 0, 662))
        self.assertEqual(histogram("F"), (256, 0, 662))
        self.assertEqual(histogram("P"), (256, 0, 1871))
        self.assertEqual(histogram("RGB"), (768, 4, 675))
        self.assertEqual(histogram("RGBA"), (1024, 0, 16384))
        self.assertEqual(histogram("CMYK"), (1024, 0, 16384))
        self.assertEqual(histogram("YCbCr"), (768, 0, 1908))



import collections
import os
import sys

import PIL

from . import Image

modules = {
    "pil": "PIL._imaging",
    "tkinter": "PIL._tkinter_finder",
    "freetype2": "PIL._imagingft",
    "littlecms2": "PIL._imagingcms",
    "webp": "PIL._webp",
}


def check_module(feature):
    if not (feature in modules):
        raise ValueError("Unknown module %s" % feature)

    module = modules[feature]

    try:
        __import__(module)
        return True
    except ImportError:
        return False


def get_supported_modules():
    return [f for f in modules if check_module(f)]


codecs = {"jpg": "jpeg", "jpg_2000": "jpeg2k", "zlib": "zip", "libtiff": "libtiff"}


def check_codec(feature):
    if feature not in codecs:
        raise ValueError("Unknown codec %s" % feature)

    codec = codecs[feature]

    return codec + "_encoder" in dir(Image.core)


def get_supported_codecs():
    return [f for f in codecs if check_codec(f)]


features = {
    "webp_anim": ("PIL._webp", "HAVE_WEBPANIM"),
    "webp_mux": ("PIL._webp", "HAVE_WEBPMUX"),
    "transp_webp": ("PIL._webp", "HAVE_TRANSPARENCY"),
    "raqm": ("PIL._imagingft", "HAVE_RAQM"),
    "libjpeg_turbo": ("PIL._imaging", "HAVE_LIBJPEGTURBO"),
    "libimagequant": ("PIL._imaging", "HAVE_LIBIMAGEQUANT"),
}


def check_feature(feature):
    if feature not in features:
        raise ValueError("Unknown feature %s" % feature)

    module, flag = features[feature]

    try:
        imported_module = __import__(module, fromlist=["PIL"])
        return getattr(imported_module, flag)
    except ImportError:
        return None


def get_supported_features():
    return [f for f in features if check_feature(f)]


def check(feature):
    return (
        feature in modules
        and check_module(feature)
        or feature in codecs
        and check_codec(feature)
        or feature in features
        and check_feature(feature)
    )


def get_supported():
    ret = get_supported_modules()
    ret.extend(get_supported_features())
    ret.extend(get_supported_codecs())
    return ret


def pilinfo(out=None, supported_formats=True):
    if out is None:
        out = sys.stdout

    Image.init()

    print("-" * 68, file=out)
    print("Pillow {}".format(PIL.__version__), file=out)
    py_version = sys.version.splitlines()
    print("Python {}".format(py_version[0].strip()), file=out)
    for py_version in py_version[1:]:
        print("       {}".format(py_version.strip()), file=out)
    print("-" * 68, file=out)
    print(
        "Python modules loaded from {}".format(os.path.dirname(Image.__file__)),
        file=out,
    )
    print(
        "Binary modules loaded from {}".format(os.path.dirname(Image.core.__file__)),
        file=out,
    )
    print("-" * 68, file=out)

    for name, feature in [
        ("pil", "PIL CORE"),
        ("tkinter", "TKINTER"),
        ("freetype2", "FREETYPE2"),
        ("littlecms2", "LITTLECMS2"),
        ("webp", "WEBP"),
        ("transp_webp", "WEBP Transparency"),
        ("webp_mux", "WEBPMUX"),
        ("webp_anim", "WEBP Animation"),
        ("jpg", "JPEG"),
        ("jpg_2000", "OPENJPEG (JPEG2000)"),
        ("zlib", "ZLIB (PNG/ZIP)"),
        ("libtiff", "LIBTIFF"),
        ("raqm", "RAQM (Bidirectional Text)"),
        ("libimagequant", "LIBIMAGEQUANT (Quantization method)"),
    ]:
        if check(name):
            print("---", feature, "support ok", file=out)
        else:
            print("***", feature, "support not installed", file=out)
    print("-" * 68, file=out)

    if supported_formats:
        extensions = collections.defaultdict(list)
        for ext, i in Image.EXTENSION.items():
            extensions[i].append(ext)

        for i in sorted(Image.ID):
            line = "{}".format(i)
            if i in Image.MIME:
                line = "{} {}".format(line, Image.MIME[i])
            print(line, file=out)

            if i in extensions:
                print(
                    "Extensions: {}".format(", ".join(sorted(extensions[i]))), file=out
                )

            features = []
            if i in Image.OPEN:
                features.append("open")
            if i in Image.SAVE:
                features.append("save")
            if i in Image.SAVE_ALL:
                features.append("save_all")
            if i in Image.DECODERS:
                features.append("decode")
            if i in Image.ENCODERS:
                features.append("encode")

            print("Features: {}".format(", ".join(features)), file=out)
            print("-" * 68, file=out)



import os.path
import unittest

from PIL import Image, ImageDraw2, features

from .helper import PillowTestCase, hopper

BLACK = (0, 0, 0)
WHITE = (255, 255, 255)
GRAY = (190, 190, 190)
DEFAULT_MODE = "RGB"
IMAGES_PATH = os.path.join("Tests", "images", "imagedraw")

# Image size
W, H = 100, 100

# Bounding box points
X0 = int(W / 4)
X1 = int(X0 * 3)
Y0 = int(H / 4)
Y1 = int(X0 * 3)

# Two kinds of bounding box
BBOX1 = [(X0, Y0), (X1, Y1)]
BBOX2 = [X0, Y0, X1, Y1]

# Two kinds of coordinate sequences
POINTS1 = [(10, 10), (20, 40), (30, 30)]
POINTS2 = [10, 10, 20, 40, 30, 30]

KITE_POINTS = [(10, 50), (70, 10), (90, 50), (70, 90), (10, 50)]

HAS_FREETYPE = features.check("freetype2")
FONT_PATH = "Tests/fonts/FreeMono.ttf"


class TestImageDraw(PillowTestCase):
    def test_sanity(self):
        im = hopper("RGB").copy()

        draw = ImageDraw2.Draw(im)
        pen = ImageDraw2.Pen("blue", width=7)
        draw.line(list(range(10)), pen)

        from PIL import ImageDraw

        draw, handler = ImageDraw.getdraw(im)
        pen = ImageDraw2.Pen("blue", width=7)
        draw.line(list(range(10)), pen)

    def helper_ellipse(self, mode, bbox):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw2.Draw(im)
        pen = ImageDraw2.Pen("blue", width=2)
        brush = ImageDraw2.Brush("green")
        expected = "Tests/images/imagedraw_ellipse_{}.png".format(mode)

        # Act
        draw.ellipse(bbox, pen, brush)

        # Assert
        self.assert_image_similar(im, Image.open(expected), 1)

    def test_ellipse1(self):
        self.helper_ellipse("RGB", BBOX1)

    def test_ellipse2(self):
        self.helper_ellipse("RGB", BBOX2)

    def test_ellipse_edge(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw2.Draw(im)
        brush = ImageDraw2.Brush("white")

        # Act
        draw.ellipse(((0, 0), (W - 1, H)), brush)

        # Assert
        self.assert_image_similar(
            im, Image.open("Tests/images/imagedraw_ellipse_edge.png"), 1
        )

    def helper_line(self, points):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw2.Draw(im)
        pen = ImageDraw2.Pen("yellow", width=2)

        # Act
        draw.line(points, pen)

        # Assert
        self.assert_image_equal(im, Image.open("Tests/images/imagedraw_line.png"))

    def test_line1_pen(self):
        self.helper_line(POINTS1)

    def test_line2_pen(self):
        self.helper_line(POINTS2)

    def test_line_pen_as_brush(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw2.Draw(im)
        pen = None
        brush = ImageDraw2.Pen("yellow", width=2)

        # Act
        # Pass in the pen as the brush parameter
        draw.line(POINTS1, pen, brush)

        # Assert
        self.assert_image_equal(im, Image.open("Tests/images/imagedraw_line.png"))

    def helper_polygon(self, points):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw2.Draw(im)
        pen = ImageDraw2.Pen("blue", width=2)
        brush = ImageDraw2.Brush("red")

        # Act
        draw.polygon(points, pen, brush)

        # Assert
        self.assert_image_equal(im, Image.open("Tests/images/imagedraw_polygon.png"))

    def test_polygon1(self):
        self.helper_polygon(POINTS1)

    def test_polygon2(self):
        self.helper_polygon(POINTS2)

    def helper_rectangle(self, bbox):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw2.Draw(im)
        pen = ImageDraw2.Pen("green", width=2)
        brush = ImageDraw2.Brush("black")

        # Act
        draw.rectangle(bbox, pen, brush)

        # Assert
        self.assert_image_equal(im, Image.open("Tests/images/imagedraw_rectangle.png"))

    def test_rectangle1(self):
        self.helper_rectangle(BBOX1)

    def test_rectangle2(self):
        self.helper_rectangle(BBOX2)

    def test_big_rectangle(self):
        # Test drawing a rectangle bigger than the image
        # Arrange
        im = Image.new("RGB", (W, H))
        bbox = [(-1, -1), (W + 1, H + 1)]
        brush = ImageDraw2.Brush("orange")
        draw = ImageDraw2.Draw(im)
        expected = "Tests/images/imagedraw_big_rectangle.png"

        # Act
        draw.rectangle(bbox, brush)

        # Assert
        self.assert_image_similar(im, Image.open(expected), 1)

    @unittest.skipUnless(HAS_FREETYPE, "ImageFont not available")
    def test_text(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw2.Draw(im)
        font = ImageDraw2.Font("white", FONT_PATH)
        expected = "Tests/images/imagedraw2_text.png"

        # Act
        draw.text((5, 5), "ImageDraw2", font)

        # Assert
        self.assert_image_similar(im, Image.open(expected), 13)

    @unittest.skipUnless(HAS_FREETYPE, "ImageFont not available")
    def test_textsize(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw2.Draw(im)
        font = ImageDraw2.Font("white", FONT_PATH)

        # Act
        size = draw.textsize("ImageDraw2", font)

        # Assert
        self.assertEqual(size[1], 12)

    @unittest.skipUnless(HAS_FREETYPE, "ImageFont not available")
    def test_textsize_empty_string(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw2.Draw(im)
        font = ImageDraw2.Font("white", FONT_PATH)

        # Act
        # Should not cause 'SystemError: <built-in method getsize of
        # ImagingFont object at 0x...> returned NULL without setting an error'
        draw.textsize("", font)
        draw.textsize("\n", font)
        draw.textsize("test\n", font)

    @unittest.skipUnless(HAS_FREETYPE, "ImageFont not available")
    def test_flush(self):
        # Arrange
        im = Image.new("RGB", (W, H))
        draw = ImageDraw2.Draw(im)
        font = ImageDraw2.Font("white", FONT_PATH)

        # Act
        draw.text((5, 5), "ImageDraw2", font)
        im2 = draw.flush()

        # Assert
        self.assert_image_equal(im, im2)



def pytest_report_header(config):
    import io

    try:
        from PIL import features

        with io.StringIO() as out:
            features.pilinfo(out=out, supported_formats=False)
            return out.getvalue()
    except Exception as e:
        return "pytest_report_header failed: %s" % str(e)



"""
Tests for resize functionality.
"""
from itertools import permutations

from PIL import Image

from .helper import PillowTestCase, hopper


class TestImagingCoreResize(PillowTestCase):
    def resize(self, im, size, f):
        # Image class independent version of resize.
        im.load()
        return im._new(im.im.resize(size, f))

    def test_nearest_mode(self):
        for mode in [
            "1",
            "P",
            "L",
            "I",
            "F",
            "RGB",
            "RGBA",
            "CMYK",
            "YCbCr",
            "I;16",
        ]:  # exotic mode
            im = hopper(mode)
            r = self.resize(im, (15, 12), Image.NEAREST)
            self.assertEqual(r.mode, mode)
            self.assertEqual(r.size, (15, 12))
            self.assertEqual(r.im.bands, im.im.bands)

    def test_convolution_modes(self):
        self.assertRaises(
            ValueError, self.resize, hopper("1"), (15, 12), Image.BILINEAR
        )
        self.assertRaises(
            ValueError, self.resize, hopper("P"), (15, 12), Image.BILINEAR
        )
        self.assertRaises(
            ValueError, self.resize, hopper("I;16"), (15, 12), Image.BILINEAR
        )
        for mode in ["L", "I", "F", "RGB", "RGBA", "CMYK", "YCbCr"]:
            im = hopper(mode)
            r = self.resize(im, (15, 12), Image.BILINEAR)
            self.assertEqual(r.mode, mode)
            self.assertEqual(r.size, (15, 12))
            self.assertEqual(r.im.bands, im.im.bands)

    def test_reduce_filters(self):
        for f in [
            Image.NEAREST,
            Image.BOX,
            Image.BILINEAR,
            Image.HAMMING,
            Image.BICUBIC,
            Image.LANCZOS,
        ]:
            r = self.resize(hopper("RGB"), (15, 12), f)
            self.assertEqual(r.mode, "RGB")
            self.assertEqual(r.size, (15, 12))

    def test_enlarge_filters(self):
        for f in [
            Image.NEAREST,
            Image.BOX,
            Image.BILINEAR,
            Image.HAMMING,
            Image.BICUBIC,
            Image.LANCZOS,
        ]:
            r = self.resize(hopper("RGB"), (212, 195), f)
            self.assertEqual(r.mode, "RGB")
            self.assertEqual(r.size, (212, 195))

    def test_endianness(self):
        # Make an image with one colored pixel, in one channel.
        # When resized, that channel should be the same as a GS image.
        # Other channels should be unaffected.
        # The R and A channels should not swap, which is indicative of
        # an endianness issues.

        samples = {
            "blank": Image.new("L", (2, 2), 0),
            "filled": Image.new("L", (2, 2), 255),
            "dirty": Image.new("L", (2, 2), 0),
        }
        samples["dirty"].putpixel((1, 1), 128)

        for f in [
            Image.NEAREST,
            Image.BOX,
            Image.BILINEAR,
            Image.HAMMING,
            Image.BICUBIC,
            Image.LANCZOS,
        ]:
            # samples resized with current filter
            references = {
                name: self.resize(ch, (4, 4), f) for name, ch in samples.items()
            }

            for mode, channels_set in [
                ("RGB", ("blank", "filled", "dirty")),
                ("RGBA", ("blank", "blank", "filled", "dirty")),
                ("LA", ("filled", "dirty")),
            ]:
                for channels in set(permutations(channels_set)):
                    # compile image from different channels permutations
                    im = Image.merge(mode, [samples[ch] for ch in channels])
                    resized = self.resize(im, (4, 4), f)

                    for i, ch in enumerate(resized.split()):
                        # check what resized channel in image is the same
                        # as separately resized channel
                        self.assert_image_equal(ch, references[channels[i]])

    def test_enlarge_zero(self):
        for f in [
            Image.NEAREST,
            Image.BOX,
            Image.BILINEAR,
            Image.HAMMING,
            Image.BICUBIC,
            Image.LANCZOS,
        ]:
            r = self.resize(Image.new("RGB", (0, 0), "white"), (212, 195), f)
            self.assertEqual(r.mode, "RGB")
            self.assertEqual(r.size, (212, 195))
            self.assertEqual(r.getdata()[0], (0, 0, 0))

    def test_unknown_filter(self):
        self.assertRaises(ValueError, self.resize, hopper(), (10, 10), 9)


class TestReducingGapResize(PillowTestCase):
    @classmethod
    def setUpClass(cls):
        cls.gradients_image = Image.open("Tests/images/radial_gradients.png")
        cls.gradients_image.load()

    def test_reducing_gap_values(self):
        ref = self.gradients_image.resize((52, 34), Image.BICUBIC, reducing_gap=None)
        im = self.gradients_image.resize((52, 34), Image.BICUBIC)
        self.assert_image_equal(ref, im)

        with self.assertRaises(ValueError):
            self.gradients_image.resize((52, 34), Image.BICUBIC, reducing_gap=0)

        with self.assertRaises(ValueError):
            self.gradients_image.resize((52, 34), Image.BICUBIC, reducing_gap=0.99)

    def test_reducing_gap_1(self):
        for box, epsilon in [
            (None, 4),
            ((1.1, 2.2, 510.8, 510.9), 4),
            ((3, 10, 410, 256), 10),
        ]:
            ref = self.gradients_image.resize((52, 34), Image.BICUBIC, box=box)
            im = self.gradients_image.resize(
                (52, 34), Image.BICUBIC, box=box, reducing_gap=1.0
            )

            with self.assertRaises(AssertionError):
                self.assert_image_equal(ref, im)

            self.assert_image_similar(ref, im, epsilon)

    def test_reducing_gap_2(self):
        for box, epsilon in [
            (None, 1.5),
            ((1.1, 2.2, 510.8, 510.9), 1.5),
            ((3, 10, 410, 256), 1),
        ]:
            ref = self.gradients_image.resize((52, 34), Image.BICUBIC, box=box)
            im = self.gradients_image.resize(
                (52, 34), Image.BICUBIC, box=box, reducing_gap=2.0
            )

            with self.assertRaises(AssertionError):
                self.assert_image_equal(ref, im)

            self.assert_image_similar(ref, im, epsilon)

    def test_reducing_gap_3(self):
        for box, epsilon in [
            (None, 1),
            ((1.1, 2.2, 510.8, 510.9), 1),
            ((3, 10, 410, 256), 0.5),
        ]:
            ref = self.gradients_image.resize((52, 34), Image.BICUBIC, box=box)
            im = self.gradients_image.resize(
                (52, 34), Image.BICUBIC, box=box, reducing_gap=3.0
            )

            with self.assertRaises(AssertionError):
                self.assert_image_equal(ref, im)

            self.assert_image_similar(ref, im, epsilon)

    def test_reducing_gap_8(self):
        for box in [None, (1.1, 2.2, 510.8, 510.9), (3, 10, 410, 256)]:
            ref = self.gradients_image.resize((52, 34), Image.BICUBIC, box=box)
            im = self.gradients_image.resize(
                (52, 34), Image.BICUBIC, box=box, reducing_gap=8.0
            )

            self.assert_image_equal(ref, im)

    def test_box_filter(self):
        for box, epsilon in [
            ((0, 0, 512, 512), 5.5),
            ((0.9, 1.7, 128, 128), 9.5),
        ]:
            ref = self.gradients_image.resize((52, 34), Image.BOX, box=box)
            im = self.gradients_image.resize(
                (52, 34), Image.BOX, box=box, reducing_gap=1.0
            )

            self.assert_image_similar(ref, im, epsilon)


class TestImageResize(PillowTestCase):
    def test_resize(self):
        def resize(mode, size):
            out = hopper(mode).resize(size)
            self.assertEqual(out.mode, mode)
            self.assertEqual(out.size, size)

        for mode in "1", "P", "L", "RGB", "I", "F":
            resize(mode, (112, 103))
            resize(mode, (188, 214))

        # Test unknown resampling filter
        with hopper() as im:
            self.assertRaises(ValueError, im.resize, (10, 10), "unknown")

    def test_default_filter(self):
        for mode in "L", "RGB", "I", "F":
            im = hopper(mode)
            self.assertEqual(im.resize((20, 20), Image.BICUBIC), im.resize((20, 20)))

        for mode in "1", "P":
            im = hopper(mode)
            self.assertEqual(im.resize((20, 20), Image.NEAREST), im.resize((20, 20)))



import unittest

from PIL import Image, ImageShow

from .helper import PillowTestCase, hopper, is_win32, on_ci, on_github_actions


class TestImageShow(PillowTestCase):
    def test_sanity(self):
        dir(Image)
        dir(ImageShow)

    def test_register(self):
        # Test registering a viewer that is not a class
        ImageShow.register("not a class")

        # Restore original state
        ImageShow._viewers.pop()

    def test_viewer_show(self):
        class TestViewer(ImageShow.Viewer):
            methodCalled = False

            def show_image(self, image, **options):
                self.methodCalled = True
                return True

        viewer = TestViewer()
        ImageShow.register(viewer, -1)

        for mode in ("1", "I;16", "LA", "RGB", "RGBA"):
            with hopper() as im:
                self.assertTrue(ImageShow.show(im))
            self.assertTrue(viewer.methodCalled)

        # Restore original state
        ImageShow._viewers.pop(0)

    @unittest.skipUnless(
        on_ci() and not (is_win32() and on_github_actions()),
        "Only run on CIs; hangs on Windows on GitHub Actions",
    )
    def test_show(self):
        for mode in ("1", "I;16", "LA", "RGB", "RGBA"):
            im = hopper(mode)
            self.assertTrue(ImageShow.show(im))

    def test_viewer(self):
        viewer = ImageShow.Viewer()

        self.assertIsNone(viewer.get_format(None))

        self.assertRaises(NotImplementedError, viewer.get_command, None)

    def test_viewers(self):
        for viewer in ImageShow._viewers:
            viewer.get_command("test.jpg")



import unittest

from PIL import Image, ImageDraw, ImageFont, features

from .helper import PillowTestCase

FONT_SIZE = 20
FONT_PATH = "Tests/fonts/DejaVuSans.ttf"


@unittest.skipUnless(features.check("raqm"), "Raqm Library is not installed.")
class TestImagecomplextext(PillowTestCase):
    def test_english(self):
        # smoke test, this should not fail
        ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)
        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)
        draw.text((0, 0), "TEST", font=ttf, fill=500, direction="ltr")

    def test_complex_text(self):
        ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)
        draw.text((0, 0), "اهلا عمان", font=ttf, fill=500)

        target = "Tests/images/test_text.png"
        with Image.open(target) as target_img:
            self.assert_image_similar(im, target_img, 0.5)

    def test_y_offset(self):
        ttf = ImageFont.truetype("Tests/fonts/NotoNastaliqUrdu-Regular.ttf", FONT_SIZE)

        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)
        draw.text((0, 0), "العالم العربي", font=ttf, fill=500)

        target = "Tests/images/test_y_offset.png"
        with Image.open(target) as target_img:
            self.assert_image_similar(im, target_img, 1.7)

    def test_complex_unicode_text(self):
        ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)
        draw.text((0, 0), "السلام عليكم", font=ttf, fill=500)

        target = "Tests/images/test_complex_unicode_text.png"
        with Image.open(target) as target_img:
            self.assert_image_similar(im, target_img, 0.5)

        ttf = ImageFont.truetype("Tests/fonts/KhmerOSBattambang-Regular.ttf", FONT_SIZE)

        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)
        draw.text((0, 0), "លោកុប្បត្តិ", font=ttf, fill=500)

        target = "Tests/images/test_complex_unicode_text2.png"
        with Image.open(target) as target_img:
            self.assert_image_similar(im, target_img, 2.3)

    def test_text_direction_rtl(self):
        ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)
        draw.text((0, 0), "English عربي", font=ttf, fill=500, direction="rtl")

        target = "Tests/images/test_direction_rtl.png"
        with Image.open(target) as target_img:
            self.assert_image_similar(im, target_img, 0.5)

    def test_text_direction_ltr(self):
        ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)
        draw.text((0, 0), "سلطنة عمان Oman", font=ttf, fill=500, direction="ltr")

        target = "Tests/images/test_direction_ltr.png"
        with Image.open(target) as target_img:
            self.assert_image_similar(im, target_img, 0.5)

    def test_text_direction_rtl2(self):
        ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)
        draw.text((0, 0), "Oman سلطنة عمان", font=ttf, fill=500, direction="rtl")

        target = "Tests/images/test_direction_ltr.png"
        with Image.open(target) as target_img:
            self.assert_image_similar(im, target_img, 0.5)

    def test_text_direction_ttb(self):
        ttf = ImageFont.truetype("Tests/fonts/NotoSansJP-Regular.otf", FONT_SIZE)

        im = Image.new(mode="RGB", size=(100, 300))
        draw = ImageDraw.Draw(im)
        try:
            draw.text((0, 0), "English あい", font=ttf, fill=500, direction="ttb")
        except ValueError as ex:
            if str(ex) == "libraqm 0.7 or greater required for 'ttb' direction":
                self.skipTest("libraqm 0.7 or greater not available")

        target = "Tests/images/test_direction_ttb.png"
        with Image.open(target) as target_img:
            self.assert_image_similar(im, target_img, 1.15)

    def test_text_direction_ttb_stroke(self):
        ttf = ImageFont.truetype("Tests/fonts/NotoSansJP-Regular.otf", 50)

        im = Image.new(mode="RGB", size=(100, 300))
        draw = ImageDraw.Draw(im)
        try:
            draw.text(
                (25, 25),
                "あい",
                font=ttf,
                fill=500,
                direction="ttb",
                stroke_width=2,
                stroke_fill="#0f0",
            )
        except ValueError as ex:
            if str(ex) == "libraqm 0.7 or greater required for 'ttb' direction":
                self.skipTest("libraqm 0.7 or greater not available")

        target = "Tests/images/test_direction_ttb_stroke.png"
        with Image.open(target) as target_img:
            self.assert_image_similar(im, target_img, 12.4)

    def test_ligature_features(self):
        ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)
        draw.text((0, 0), "filling", font=ttf, fill=500, features=["-liga"])
        target = "Tests/images/test_ligature_features.png"
        with Image.open(target) as target_img:
            self.assert_image_similar(im, target_img, 0.5)

        liga_size = ttf.getsize("fi", features=["-liga"])
        self.assertEqual(liga_size, (13, 19))

    def test_kerning_features(self):
        ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)
        draw.text((0, 0), "TeToAV", font=ttf, fill=500, features=["-kern"])

        target = "Tests/images/test_kerning_features.png"
        with Image.open(target) as target_img:
            self.assert_image_similar(im, target_img, 0.5)

    def test_arabictext_features(self):
        ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)
        draw.text(
            (0, 0),
            "اللغة العربية",
            font=ttf,
            fill=500,
            features=["-fina", "-init", "-medi"],
        )

        target = "Tests/images/test_arabictext_features.png"
        with Image.open(target) as target_img:
            self.assert_image_similar(im, target_img, 0.5)

    def test_x_max_and_y_offset(self):
        ttf = ImageFont.truetype("Tests/fonts/ArefRuqaa-Regular.ttf", 40)

        im = Image.new(mode="RGB", size=(50, 100))
        draw = ImageDraw.Draw(im)
        draw.text((0, 0), "لح", font=ttf, fill=500)

        target = "Tests/images/test_x_max_and_y_offset.png"
        with Image.open(target) as target_img:
            self.assert_image_similar(im, target_img, 0.5)

    def test_language(self):
        ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)
        draw.text((0, 0), "абвг", font=ttf, fill=500, language="sr")

        target = "Tests/images/test_language.png"
        with Image.open(target) as target_img:
            self.assert_image_similar(im, target_img, 0.5)



from PIL import Image, ImageQt

from .helper import PillowTestCase, hopper
from .test_imageqt import PillowQtTestCase


class TestFromQImage(PillowQtTestCase, PillowTestCase):
    def setUp(self):
        super().setUp()
        self.files_to_test = [
            hopper(),
            Image.open("Tests/images/transparent.png"),
            Image.open("Tests/images/7x13.png"),
        ]
        for im in self.files_to_test:
            self.addCleanup(im.close)

    def roundtrip(self, expected):
        # PIL -> Qt
        intermediate = expected.toqimage()
        # Qt -> PIL
        result = ImageQt.fromqimage(intermediate)

        if intermediate.hasAlphaChannel():
            self.assert_image_equal(result, expected.convert("RGBA"))
        else:
            self.assert_image_equal(result, expected.convert("RGB"))

    def test_sanity_1(self):
        for im in self.files_to_test:
            self.roundtrip(im.convert("1"))

    def test_sanity_rgb(self):
        for im in self.files_to_test:
            self.roundtrip(im.convert("RGB"))

    def test_sanity_rgba(self):
        for im in self.files_to_test:
            self.roundtrip(im.convert("RGBA"))

    def test_sanity_l(self):
        for im in self.files_to_test:
            self.roundtrip(im.convert("L"))

    def test_sanity_p(self):
        for im in self.files_to_test:
            self.roundtrip(im.convert("P"))



from PIL import _binary

from .helper import PillowTestCase


class TestBinary(PillowTestCase):
    def test_standard(self):
        self.assertEqual(_binary.i8(b"*"), 42)
        self.assertEqual(_binary.o8(42), b"*")

    def test_little_endian(self):
        self.assertEqual(_binary.i16le(b"\xff\xff\x00\x00"), 65535)
        self.assertEqual(_binary.i32le(b"\xff\xff\x00\x00"), 65535)

        self.assertEqual(_binary.o16le(65535), b"\xff\xff")
        self.assertEqual(_binary.o32le(65535), b"\xff\xff\x00\x00")

    def test_big_endian(self):
        self.assertEqual(_binary.i16be(b"\x00\x00\xff\xff"), 0)
        self.assertEqual(_binary.i32be(b"\x00\x00\xff\xff"), 65535)

        self.assertEqual(_binary.o16be(65535), b"\xff\xff")
        self.assertEqual(_binary.o32be(65535), b"\x00\x00\xff\xff")



import ctypes
from io import BytesIO

from PIL import Image, ImageWin

from .helper import PillowTestCase, hopper, is_win32

# see https://github.com/python-pillow/Pillow/pull/1431#issuecomment-144692652

if is_win32():
    import ctypes.wintypes

    class BITMAPFILEHEADER(ctypes.Structure):
        _pack_ = 2
        _fields_ = [
            ("bfType", ctypes.wintypes.WORD),
            ("bfSize", ctypes.wintypes.DWORD),
            ("bfReserved1", ctypes.wintypes.WORD),
            ("bfReserved2", ctypes.wintypes.WORD),
            ("bfOffBits", ctypes.wintypes.DWORD),
        ]

    class BITMAPINFOHEADER(ctypes.Structure):
        _pack_ = 2
        _fields_ = [
            ("biSize", ctypes.wintypes.DWORD),
            ("biWidth", ctypes.wintypes.LONG),
            ("biHeight", ctypes.wintypes.LONG),
            ("biPlanes", ctypes.wintypes.WORD),
            ("biBitCount", ctypes.wintypes.WORD),
            ("biCompression", ctypes.wintypes.DWORD),
            ("biSizeImage", ctypes.wintypes.DWORD),
            ("biXPelsPerMeter", ctypes.wintypes.LONG),
            ("biYPelsPerMeter", ctypes.wintypes.LONG),
            ("biClrUsed", ctypes.wintypes.DWORD),
            ("biClrImportant", ctypes.wintypes.DWORD),
        ]

    BI_RGB = 0
    DIB_RGB_COLORS = 0

    memcpy = ctypes.cdll.msvcrt.memcpy
    memcpy.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t]

    CreateCompatibleDC = ctypes.windll.gdi32.CreateCompatibleDC
    CreateCompatibleDC.argtypes = [ctypes.wintypes.HDC]
    CreateCompatibleDC.restype = ctypes.wintypes.HDC

    DeleteDC = ctypes.windll.gdi32.DeleteDC
    DeleteDC.argtypes = [ctypes.wintypes.HDC]

    SelectObject = ctypes.windll.gdi32.SelectObject
    SelectObject.argtypes = [ctypes.wintypes.HDC, ctypes.wintypes.HGDIOBJ]
    SelectObject.restype = ctypes.wintypes.HGDIOBJ

    DeleteObject = ctypes.windll.gdi32.DeleteObject
    DeleteObject.argtypes = [ctypes.wintypes.HGDIOBJ]

    CreateDIBSection = ctypes.windll.gdi32.CreateDIBSection
    CreateDIBSection.argtypes = [
        ctypes.wintypes.HDC,
        ctypes.c_void_p,
        ctypes.c_uint,
        ctypes.POINTER(ctypes.c_void_p),
        ctypes.wintypes.HANDLE,
        ctypes.wintypes.DWORD,
    ]
    CreateDIBSection.restype = ctypes.wintypes.HBITMAP

    def serialize_dib(bi, pixels):
        bf = BITMAPFILEHEADER()
        bf.bfType = 0x4D42
        bf.bfOffBits = ctypes.sizeof(bf) + bi.biSize
        bf.bfSize = bf.bfOffBits + bi.biSizeImage
        bf.bfReserved1 = bf.bfReserved2 = 0

        buf = (ctypes.c_byte * bf.bfSize)()
        bp = ctypes.addressof(buf)
        memcpy(bp, ctypes.byref(bf), ctypes.sizeof(bf))
        memcpy(bp + ctypes.sizeof(bf), ctypes.byref(bi), bi.biSize)
        memcpy(bp + bf.bfOffBits, pixels, bi.biSizeImage)
        return bytearray(buf)

    class TestImageWinPointers(PillowTestCase):
        def test_pointer(self):
            im = hopper()
            (width, height) = im.size
            opath = self.tempfile("temp.png")
            imdib = ImageWin.Dib(im)

            hdr = BITMAPINFOHEADER()
            hdr.biSize = ctypes.sizeof(hdr)
            hdr.biWidth = width
            hdr.biHeight = height
            hdr.biPlanes = 1
            hdr.biBitCount = 32
            hdr.biCompression = BI_RGB
            hdr.biSizeImage = width * height * 4
            hdr.biClrUsed = 0
            hdr.biClrImportant = 0

            hdc = CreateCompatibleDC(None)
            pixels = ctypes.c_void_p()
            dib = CreateDIBSection(
                hdc, ctypes.byref(hdr), DIB_RGB_COLORS, ctypes.byref(pixels), None, 0
            )
            SelectObject(hdc, dib)

            imdib.expose(hdc)
            bitmap = serialize_dib(hdr, pixels)
            DeleteObject(dib)
            DeleteDC(hdc)

            Image.open(BytesIO(bitmap)).save(opath)



from PIL import Image, ImageColor

from .helper import PillowTestCase


class TestImageColor(PillowTestCase):
    def test_hash(self):
        # short 3 components
        self.assertEqual((255, 0, 0), ImageColor.getrgb("#f00"))
        self.assertEqual((0, 255, 0), ImageColor.getrgb("#0f0"))
        self.assertEqual((0, 0, 255), ImageColor.getrgb("#00f"))

        # short 4 components
        self.assertEqual((255, 0, 0, 0), ImageColor.getrgb("#f000"))
        self.assertEqual((0, 255, 0, 0), ImageColor.getrgb("#0f00"))
        self.assertEqual((0, 0, 255, 0), ImageColor.getrgb("#00f0"))
        self.assertEqual((0, 0, 0, 255), ImageColor.getrgb("#000f"))

        # long 3 components
        self.assertEqual((222, 0, 0), ImageColor.getrgb("#de0000"))
        self.assertEqual((0, 222, 0), ImageColor.getrgb("#00de00"))
        self.assertEqual((0, 0, 222), ImageColor.getrgb("#0000de"))

        # long 4 components
        self.assertEqual((222, 0, 0, 0), ImageColor.getrgb("#de000000"))
        self.assertEqual((0, 222, 0, 0), ImageColor.getrgb("#00de0000"))
        self.assertEqual((0, 0, 222, 0), ImageColor.getrgb("#0000de00"))
        self.assertEqual((0, 0, 0, 222), ImageColor.getrgb("#000000de"))

        # case insensitivity
        self.assertEqual(ImageColor.getrgb("#DEF"), ImageColor.getrgb("#def"))
        self.assertEqual(ImageColor.getrgb("#CDEF"), ImageColor.getrgb("#cdef"))
        self.assertEqual(ImageColor.getrgb("#DEFDEF"), ImageColor.getrgb("#defdef"))
        self.assertEqual(ImageColor.getrgb("#CDEFCDEF"), ImageColor.getrgb("#cdefcdef"))

        # not a number
        self.assertRaises(ValueError, ImageColor.getrgb, "#fo0")
        self.assertRaises(ValueError, ImageColor.getrgb, "#fo00")
        self.assertRaises(ValueError, ImageColor.getrgb, "#fo0000")
        self.assertRaises(ValueError, ImageColor.getrgb, "#fo000000")

        # wrong number of components
        self.assertRaises(ValueError, ImageColor.getrgb, "#f0000")
        self.assertRaises(ValueError, ImageColor.getrgb, "#f000000")
        self.assertRaises(ValueError, ImageColor.getrgb, "#f00000000")
        self.assertRaises(ValueError, ImageColor.getrgb, "#f000000000")
        self.assertRaises(ValueError, ImageColor.getrgb, "#f00000 ")

    def test_colormap(self):
        self.assertEqual((0, 0, 0), ImageColor.getrgb("black"))
        self.assertEqual((255, 255, 255), ImageColor.getrgb("white"))
        self.assertEqual((255, 255, 255), ImageColor.getrgb("WHITE"))

        self.assertRaises(ValueError, ImageColor.getrgb, "black ")

    def test_functions(self):
        # rgb numbers
        self.assertEqual((255, 0, 0), ImageColor.getrgb("rgb(255,0,0)"))
        self.assertEqual((0, 255, 0), ImageColor.getrgb("rgb(0,255,0)"))
        self.assertEqual((0, 0, 255), ImageColor.getrgb("rgb(0,0,255)"))

        # percents
        self.assertEqual((255, 0, 0), ImageColor.getrgb("rgb(100%,0%,0%)"))
        self.assertEqual((0, 255, 0), ImageColor.getrgb("rgb(0%,100%,0%)"))
        self.assertEqual((0, 0, 255), ImageColor.getrgb("rgb(0%,0%,100%)"))

        # rgba numbers
        self.assertEqual((255, 0, 0, 0), ImageColor.getrgb("rgba(255,0,0,0)"))
        self.assertEqual((0, 255, 0, 0), ImageColor.getrgb("rgba(0,255,0,0)"))
        self.assertEqual((0, 0, 255, 0), ImageColor.getrgb("rgba(0,0,255,0)"))
        self.assertEqual((0, 0, 0, 255), ImageColor.getrgb("rgba(0,0,0,255)"))

        self.assertEqual((255, 0, 0), ImageColor.getrgb("hsl(0,100%,50%)"))
        self.assertEqual((255, 0, 0), ImageColor.getrgb("hsl(360,100%,50%)"))
        self.assertEqual((0, 255, 255), ImageColor.getrgb("hsl(180,100%,50%)"))

        self.assertEqual((255, 0, 0), ImageColor.getrgb("hsv(0,100%,100%)"))
        self.assertEqual((255, 0, 0), ImageColor.getrgb("hsv(360,100%,100%)"))
        self.assertEqual((0, 255, 255), ImageColor.getrgb("hsv(180,100%,100%)"))

        # alternate format
        self.assertEqual(
            ImageColor.getrgb("hsb(0,100%,50%)"), ImageColor.getrgb("hsv(0,100%,50%)")
        )

        # floats
        self.assertEqual((254, 3, 3), ImageColor.getrgb("hsl(0.1,99.2%,50.3%)"))
        self.assertEqual((255, 0, 0), ImageColor.getrgb("hsl(360.,100.0%,50%)"))

        self.assertEqual((253, 2, 2), ImageColor.getrgb("hsv(0.1,99.2%,99.3%)"))
        self.assertEqual((255, 0, 0), ImageColor.getrgb("hsv(360.,100.0%,100%)"))

        # case insensitivity
        self.assertEqual(
            ImageColor.getrgb("RGB(255,0,0)"), ImageColor.getrgb("rgb(255,0,0)")
        )
        self.assertEqual(
            ImageColor.getrgb("RGB(100%,0%,0%)"), ImageColor.getrgb("rgb(100%,0%,0%)")
        )
        self.assertEqual(
            ImageColor.getrgb("RGBA(255,0,0,0)"), ImageColor.getrgb("rgba(255,0,0,0)")
        )
        self.assertEqual(
            ImageColor.getrgb("HSL(0,100%,50%)"), ImageColor.getrgb("hsl(0,100%,50%)")
        )
        self.assertEqual(
            ImageColor.getrgb("HSV(0,100%,50%)"), ImageColor.getrgb("hsv(0,100%,50%)")
        )
        self.assertEqual(
            ImageColor.getrgb("HSB(0,100%,50%)"), ImageColor.getrgb("hsb(0,100%,50%)")
        )

        # space agnosticism
        self.assertEqual((255, 0, 0), ImageColor.getrgb("rgb(  255  ,  0  ,  0  )"))
        self.assertEqual((255, 0, 0), ImageColor.getrgb("rgb(  100%  ,  0%  ,  0%  )"))
        self.assertEqual(
            (255, 0, 0, 0), ImageColor.getrgb("rgba(  255  ,  0  ,  0  ,  0  )")
        )
        self.assertEqual((255, 0, 0), ImageColor.getrgb("hsl(  0  ,  100%  ,  50%  )"))
        self.assertEqual((255, 0, 0), ImageColor.getrgb("hsv(  0  ,  100%  ,  100%  )"))

        # wrong number of components
        self.assertRaises(ValueError, ImageColor.getrgb, "rgb(255,0)")
        self.assertRaises(ValueError, ImageColor.getrgb, "rgb(255,0,0,0)")

        self.assertRaises(ValueError, ImageColor.getrgb, "rgb(100%,0%)")
        self.assertRaises(ValueError, ImageColor.getrgb, "rgb(100%,0%,0)")
        self.assertRaises(ValueError, ImageColor.getrgb, "rgb(100%,0%,0 %)")
        self.assertRaises(ValueError, ImageColor.getrgb, "rgb(100%,0%,0%,0%)")

        self.assertRaises(ValueError, ImageColor.getrgb, "rgba(255,0,0)")
        self.assertRaises(ValueError, ImageColor.getrgb, "rgba(255,0,0,0,0)")

        self.assertRaises(ValueError, ImageColor.getrgb, "hsl(0,100%)")
        self.assertRaises(ValueError, ImageColor.getrgb, "hsl(0,100%,0%,0%)")
        self.assertRaises(ValueError, ImageColor.getrgb, "hsl(0%,100%,50%)")
        self.assertRaises(ValueError, ImageColor.getrgb, "hsl(0,100,50%)")
        self.assertRaises(ValueError, ImageColor.getrgb, "hsl(0,100%,50)")

        self.assertRaises(ValueError, ImageColor.getrgb, "hsv(0,100%)")
        self.assertRaises(ValueError, ImageColor.getrgb, "hsv(0,100%,0%,0%)")
        self.assertRaises(ValueError, ImageColor.getrgb, "hsv(0%,100%,50%)")
        self.assertRaises(ValueError, ImageColor.getrgb, "hsv(0,100,50%)")
        self.assertRaises(ValueError, ImageColor.getrgb, "hsv(0,100%,50)")

    # look for rounding errors (based on code by Tim Hatch)
    def test_rounding_errors(self):

        for color in ImageColor.colormap:
            expected = Image.new("RGB", (1, 1), color).convert("L").getpixel((0, 0))
            actual = ImageColor.getcolor(color, "L")
            self.assertEqual(expected, actual)

        self.assertEqual(
            (0, 255, 115), ImageColor.getcolor("rgba(0, 255, 115, 33)", "RGB")
        )
        Image.new("RGB", (1, 1), "white")

        self.assertEqual((0, 0, 0, 255), ImageColor.getcolor("black", "RGBA"))
        self.assertEqual((255, 255, 255, 255), ImageColor.getcolor("white", "RGBA"))
        self.assertEqual(
            (0, 255, 115, 33), ImageColor.getcolor("rgba(0, 255, 115, 33)", "RGBA")
        )
        Image.new("RGBA", (1, 1), "white")

        self.assertEqual(0, ImageColor.getcolor("black", "L"))
        self.assertEqual(255, ImageColor.getcolor("white", "L"))
        self.assertEqual(163, ImageColor.getcolor("rgba(0, 255, 115, 33)", "L"))
        Image.new("L", (1, 1), "white")

        self.assertEqual(0, ImageColor.getcolor("black", "1"))
        self.assertEqual(255, ImageColor.getcolor("white", "1"))
        # The following test is wrong, but is current behavior
        # The correct result should be 255 due to the mode 1
        self.assertEqual(163, ImageColor.getcolor("rgba(0, 255, 115, 33)", "1"))
        # Correct behavior
        # self.assertEqual(
        #     255, ImageColor.getcolor("rgba(0, 255, 115, 33)", "1"))
        Image.new("1", (1, 1), "white")

        self.assertEqual((0, 255), ImageColor.getcolor("black", "LA"))
        self.assertEqual((255, 255), ImageColor.getcolor("white", "LA"))
        self.assertEqual((163, 33), ImageColor.getcolor("rgba(0, 255, 115, 33)", "LA"))
        Image.new("LA", (1, 1), "white")



from PIL import Image, ImageMath, ImageMode

from .helper import PillowTestCase, convert_to_comparable


class TestImageReduce(PillowTestCase):
    # There are several internal implementations
    remarkable_factors = [
        # special implementations
        1,
        2,
        3,
        4,
        5,
        6,
        # 1xN implementation
        (1, 2),
        (1, 3),
        (1, 4),
        (1, 7),
        # Nx1 implementation
        (2, 1),
        (3, 1),
        (4, 1),
        (7, 1),
        # general implementation with different paths
        (4, 6),
        (5, 6),
        (4, 7),
        (5, 7),
        (19, 17),
    ]

    @classmethod
    def setUpClass(cls):
        cls.gradients_image = Image.open("Tests/images/radial_gradients.png")
        cls.gradients_image.load()

    def test_args_factor(self):
        im = Image.new("L", (10, 10))

        self.assertEqual((4, 4), im.reduce(3).size)
        self.assertEqual((4, 10), im.reduce((3, 1)).size)
        self.assertEqual((10, 4), im.reduce((1, 3)).size)

        with self.assertRaises(ValueError):
            im.reduce(0)
        with self.assertRaises(TypeError):
            im.reduce(2.0)
        with self.assertRaises(ValueError):
            im.reduce((0, 10))

    def test_args_box(self):
        im = Image.new("L", (10, 10))

        self.assertEqual((5, 5), im.reduce(2, (0, 0, 10, 10)).size)
        self.assertEqual((1, 1), im.reduce(2, (5, 5, 6, 6)).size)

        with self.assertRaises(TypeError):
            im.reduce(2, "stri")
        with self.assertRaises(TypeError):
            im.reduce(2, 2)
        with self.assertRaises(ValueError):
            im.reduce(2, (0, 0, 11, 10))
        with self.assertRaises(ValueError):
            im.reduce(2, (0, 0, 10, 11))
        with self.assertRaises(ValueError):
            im.reduce(2, (-1, 0, 10, 10))
        with self.assertRaises(ValueError):
            im.reduce(2, (0, -1, 10, 10))
        with self.assertRaises(ValueError):
            im.reduce(2, (0, 5, 10, 5))
        with self.assertRaises(ValueError):
            im.reduce(2, (5, 0, 5, 10))

    def test_unsupported_modes(self):
        im = Image.new("P", (10, 10))
        with self.assertRaises(ValueError):
            im.reduce(3)

        im = Image.new("1", (10, 10))
        with self.assertRaises(ValueError):
            im.reduce(3)

        im = Image.new("I;16", (10, 10))
        with self.assertRaises(ValueError):
            im.reduce(3)

    def get_image(self, mode):
        mode_info = ImageMode.getmode(mode)
        if mode_info.basetype == "L":
            bands = [self.gradients_image]
            for _ in mode_info.bands[1:]:
                # rotate previous image
                band = bands[-1].transpose(Image.ROTATE_90)
                bands.append(band)
            # Correct alpha channel by transforming completely transparent pixels.
            # Low alpha values also emphasize error after alpha multiplication.
            if mode.endswith("A"):
                bands[-1] = bands[-1].point(lambda x: int(85 + x / 1.5))
            im = Image.merge(mode, bands)
        else:
            assert len(mode_info.bands) == 1
            im = self.gradients_image.convert(mode)
        # change the height to make a not-square image
        return im.crop((0, 0, im.width, im.height - 5))

    def compare_reduce_with_box(self, im, factor):
        box = (11, 13, 146, 164)
        reduced = im.reduce(factor, box=box)
        reference = im.crop(box).reduce(factor)
        self.assertEqual(reduced, reference)

    def compare_reduce_with_reference(self, im, factor, average_diff=0.4, max_diff=1):
        """Image.reduce() should look very similar to Image.resize(BOX).

        A reference image is compiled from a large source area
        and possible last column and last row.
        +-----------+
        |..........c|
        |..........c|
        |..........c|
        |rrrrrrrrrrp|
        +-----------+
        """
        reduced = im.reduce(factor)

        if not isinstance(factor, (list, tuple)):
            factor = (factor, factor)

        reference = Image.new(im.mode, reduced.size)
        area_size = (im.size[0] // factor[0], im.size[1] // factor[1])
        area_box = (0, 0, area_size[0] * factor[0], area_size[1] * factor[1])
        area = im.resize(area_size, Image.BOX, area_box)
        reference.paste(area, (0, 0))

        if area_size[0] < reduced.size[0]:
            self.assertEqual(reduced.size[0] - area_size[0], 1)
            last_column_box = (area_box[2], 0, im.size[0], area_box[3])
            last_column = im.resize((1, area_size[1]), Image.BOX, last_column_box)
            reference.paste(last_column, (area_size[0], 0))

        if area_size[1] < reduced.size[1]:
            self.assertEqual(reduced.size[1] - area_size[1], 1)
            last_row_box = (0, area_box[3], area_box[2], im.size[1])
            last_row = im.resize((area_size[0], 1), Image.BOX, last_row_box)
            reference.paste(last_row, (0, area_size[1]))

        if area_size[0] < reduced.size[0] and area_size[1] < reduced.size[1]:
            last_pixel_box = (area_box[2], area_box[3], im.size[0], im.size[1])
            last_pixel = im.resize((1, 1), Image.BOX, last_pixel_box)
            reference.paste(last_pixel, area_size)

        self.assert_compare_images(reduced, reference, average_diff, max_diff)

    def assert_compare_images(self, a, b, max_average_diff, max_diff=255):
        self.assertEqual(a.mode, b.mode, "got mode %r, expected %r" % (a.mode, b.mode))
        self.assertEqual(a.size, b.size, "got size %r, expected %r" % (a.size, b.size))

        a, b = convert_to_comparable(a, b)

        bands = ImageMode.getmode(a.mode).bands
        for band, ach, bch in zip(bands, a.split(), b.split()):
            ch_diff = ImageMath.eval("convert(abs(a - b), 'L')", a=ach, b=bch)
            ch_hist = ch_diff.histogram()

            average_diff = sum(i * num for i, num in enumerate(ch_hist)) / float(
                a.size[0] * a.size[1]
            )
            self.assertGreaterEqual(
                max_average_diff,
                average_diff,
                (
                    "average pixel value difference {:.4f} > expected {:.4f} "
                    "for '{}' band"
                ).format(average_diff, max_average_diff, band),
            )

            last_diff = [i for i, num in enumerate(ch_hist) if num > 0][-1]
            self.assertGreaterEqual(
                max_diff,
                last_diff,
                "max pixel value difference {} > expected {} for '{}' band".format(
                    last_diff, max_diff, band
                ),
            )

    def test_mode_L(self):
        im = self.get_image("L")
        for factor in self.remarkable_factors:
            self.compare_reduce_with_reference(im, factor)
            self.compare_reduce_with_box(im, factor)

    def test_mode_LA(self):
        im = self.get_image("LA")
        for factor in self.remarkable_factors:
            self.compare_reduce_with_reference(im, factor, 0.8, 5)

        # With opaque alpha, an error should be way smaller.
        im.putalpha(Image.new("L", im.size, 255))
        for factor in self.remarkable_factors:
            self.compare_reduce_with_reference(im, factor)
            self.compare_reduce_with_box(im, factor)

    def test_mode_La(self):
        im = self.get_image("La")
        for factor in self.remarkable_factors:
            self.compare_reduce_with_reference(im, factor)
            self.compare_reduce_with_box(im, factor)

    def test_mode_RGB(self):
        im = self.get_image("RGB")
        for factor in self.remarkable_factors:
            self.compare_reduce_with_reference(im, factor)
            self.compare_reduce_with_box(im, factor)

    def test_mode_RGBA(self):
        im = self.get_image("RGBA")
        for factor in self.remarkable_factors:
            self.compare_reduce_with_reference(im, factor, 0.8, 5)

        # With opaque alpha, an error should be way smaller.
        im.putalpha(Image.new("L", im.size, 255))
        for factor in self.remarkable_factors:
            self.compare_reduce_with_reference(im, factor)
            self.compare_reduce_with_box(im, factor)

    def test_mode_RGBa(self):
        im = self.get_image("RGBa")
        for factor in self.remarkable_factors:
            self.compare_reduce_with_reference(im, factor)
            self.compare_reduce_with_box(im, factor)

    def test_mode_I(self):
        im = self.get_image("I")
        for factor in self.remarkable_factors:
            self.compare_reduce_with_reference(im, factor)
            self.compare_reduce_with_box(im, factor)

    def test_mode_F(self):
        im = self.get_image("F")
        for factor in self.remarkable_factors:
            self.compare_reduce_with_reference(im, factor, 0, 0)
            self.compare_reduce_with_box(im, factor)



from PIL import Image

from .helper import PillowTestCase, fromstring, hopper, tostring


class TestImageThumbnail(PillowTestCase):
    def test_sanity(self):
        im = hopper()
        self.assertIsNone(im.thumbnail((100, 100)))

        self.assertEqual(im.size, (100, 100))

    def test_aspect(self):
        im = Image.new("L", (128, 128))
        im.thumbnail((100, 100))
        self.assertEqual(im.size, (100, 100))

        im = Image.new("L", (128, 256))
        im.thumbnail((100, 100))
        self.assertEqual(im.size, (50, 100))

        im = Image.new("L", (128, 256))
        im.thumbnail((50, 100))
        self.assertEqual(im.size, (50, 100))

        im = Image.new("L", (256, 128))
        im.thumbnail((100, 100))
        self.assertEqual(im.size, (100, 50))

        im = Image.new("L", (256, 128))
        im.thumbnail((100, 50))
        self.assertEqual(im.size, (100, 50))

        im = Image.new("L", (128, 128))
        im.thumbnail((100, 100))
        self.assertEqual(im.size, (100, 100))

        im = Image.new("L", (256, 162))  # ratio is 1.5802469136
        im.thumbnail((33, 33))
        self.assertEqual(im.size, (33, 21))  # ratio is 1.5714285714

        im = Image.new("L", (162, 256))  # ratio is 0.6328125
        im.thumbnail((33, 33))
        self.assertEqual(im.size, (21, 33))  # ratio is 0.6363636364

    def test_float(self):
        im = Image.new("L", (128, 128))
        im.thumbnail((99.9, 99.9))
        self.assertEqual(im.size, (100, 100))

    def test_no_resize(self):
        # Check that draft() can resize the image to the destination size
        with Image.open("Tests/images/hopper.jpg") as im:
            im.draft(None, (64, 64))
            self.assertEqual(im.size, (64, 64))

        # Test thumbnail(), where only draft() is necessary to resize the image
        with Image.open("Tests/images/hopper.jpg") as im:
            im.thumbnail((64, 64))
            self.assertEqual(im.size, (64, 64))

    def test_DCT_scaling_edges(self):
        # Make an image with red borders and size (N * 8) + 1 to cross DCT grid
        im = Image.new("RGB", (257, 257), "red")
        im.paste(Image.new("RGB", (235, 235)), (11, 11))

        thumb = fromstring(tostring(im, "JPEG", quality=99, subsampling=0))
        # small reducing_gap to amplify the effect
        thumb.thumbnail((32, 32), Image.BICUBIC, reducing_gap=1.0)

        ref = im.resize((32, 32), Image.BICUBIC)
        # This is still JPEG, some error is present. Without the fix it is 11.5
        self.assert_image_similar(thumb, ref, 1.5)

    def test_reducing_gap_values(self):
        im = hopper()
        im.thumbnail((18, 18), Image.BICUBIC)

        ref = hopper()
        ref.thumbnail((18, 18), Image.BICUBIC, reducing_gap=2.0)
        # reducing_gap=2.0 should be the default
        self.assert_image_equal(ref, im)

        ref = hopper()
        ref.thumbnail((18, 18), Image.BICUBIC, reducing_gap=None)
        with self.assertRaises(AssertionError):
            self.assert_image_equal(ref, im)

        self.assert_image_similar(ref, im, 3.5)

    def test_reducing_gap_for_DCT_scaling(self):
        with Image.open("Tests/images/hopper.jpg") as ref:
            # thumbnail should call draft with reducing_gap scale
            ref.draft(None, (18 * 3, 18 * 3))
            ref = ref.resize((18, 18), Image.BICUBIC)

            with Image.open("Tests/images/hopper.jpg") as im:
                im.thumbnail((18, 18), Image.BICUBIC, reducing_gap=3.0)

                self.assert_image_equal(ref, im)



from PIL import Image

from .helper import PillowTestCase, hopper


class TestImageConvert(PillowTestCase):
    def test_sanity(self):
        def convert(im, mode):
            out = im.convert(mode)
            self.assertEqual(out.mode, mode)
            self.assertEqual(out.size, im.size)

        modes = (
            "1",
            "L",
            "LA",
            "P",
            "PA",
            "I",
            "F",
            "RGB",
            "RGBA",
            "RGBX",
            "CMYK",
            "YCbCr",
            "HSV",
        )

        for mode in modes:
            im = hopper(mode)
            for mode in modes:
                convert(im, mode)

            # Check 0
            im = Image.new(mode, (0, 0))
            for mode in modes:
                convert(im, mode)

    def test_default(self):

        im = hopper("P")
        self.assert_image(im, "P", im.size)
        im = im.convert()
        self.assert_image(im, "RGB", im.size)
        im = im.convert()
        self.assert_image(im, "RGB", im.size)

    # ref https://github.com/python-pillow/Pillow/issues/274

    def _test_float_conversion(self, im):
        orig = im.getpixel((5, 5))
        converted = im.convert("F").getpixel((5, 5))
        self.assertEqual(orig, converted)

    def test_8bit(self):
        with Image.open("Tests/images/hopper.jpg") as im:
            self._test_float_conversion(im.convert("L"))

    def test_16bit(self):
        with Image.open("Tests/images/16bit.cropped.tif") as im:
            self._test_float_conversion(im)

    def test_16bit_workaround(self):
        with Image.open("Tests/images/16bit.cropped.tif") as im:
            self._test_float_conversion(im.convert("I"))

    def test_rgba_p(self):
        im = hopper("RGBA")
        im.putalpha(hopper("L"))

        converted = im.convert("P")
        comparable = converted.convert("RGBA")

        self.assert_image_similar(im, comparable, 20)

    def test_trns_p(self):
        im = hopper("P")
        im.info["transparency"] = 0

        f = self.tempfile("temp.png")

        im_l = im.convert("L")
        self.assertEqual(im_l.info["transparency"], 0)  # undone
        im_l.save(f)

        im_rgb = im.convert("RGB")
        self.assertEqual(im_rgb.info["transparency"], (0, 0, 0))  # undone
        im_rgb.save(f)

    # ref https://github.com/python-pillow/Pillow/issues/664

    def test_trns_p_rgba(self):
        # Arrange
        im = hopper("P")
        im.info["transparency"] = 128

        # Act
        im_rgba = im.convert("RGBA")

        # Assert
        self.assertNotIn("transparency", im_rgba.info)
        # https://github.com/python-pillow/Pillow/issues/2702
        self.assertIsNone(im_rgba.palette)

    def test_trns_l(self):
        im = hopper("L")
        im.info["transparency"] = 128

        f = self.tempfile("temp.png")

        im_rgb = im.convert("RGB")
        self.assertEqual(im_rgb.info["transparency"], (128, 128, 128))  # undone
        im_rgb.save(f)

        im_p = im.convert("P")
        self.assertIn("transparency", im_p.info)
        im_p.save(f)

        im_p = self.assert_warning(UserWarning, im.convert, "P", palette=Image.ADAPTIVE)
        self.assertNotIn("transparency", im_p.info)
        im_p.save(f)

    def test_trns_RGB(self):
        im = hopper("RGB")
        im.info["transparency"] = im.getpixel((0, 0))

        f = self.tempfile("temp.png")

        im_l = im.convert("L")
        self.assertEqual(im_l.info["transparency"], im_l.getpixel((0, 0)))  # undone
        im_l.save(f)

        im_p = im.convert("P")
        self.assertIn("transparency", im_p.info)
        im_p.save(f)

        im_rgba = im.convert("RGBA")
        self.assertNotIn("transparency", im_rgba.info)
        im_rgba.save(f)

        im_p = self.assert_warning(UserWarning, im.convert, "P", palette=Image.ADAPTIVE)
        self.assertNotIn("transparency", im_p.info)
        im_p.save(f)

    def test_gif_with_rgba_palette_to_p(self):
        # See https://github.com/python-pillow/Pillow/issues/2433
        with Image.open("Tests/images/hopper.gif") as im:
            im.info["transparency"] = 255
            im.load()
            self.assertEqual(im.palette.mode, "RGBA")
            im_p = im.convert("P")

        # Should not raise ValueError: unrecognized raw mode
        im_p.load()

    def test_p_la(self):
        im = hopper("RGBA")
        alpha = hopper("L")
        im.putalpha(alpha)

        comparable = im.convert("P").convert("LA").getchannel("A")

        self.assert_image_similar(alpha, comparable, 5)

    def test_matrix_illegal_conversion(self):
        # Arrange
        im = hopper("CMYK")
        # fmt: off
        matrix = (
            0.412453, 0.357580, 0.180423, 0,
            0.212671, 0.715160, 0.072169, 0,
            0.019334, 0.119193, 0.950227, 0)
        # fmt: on
        self.assertNotEqual(im.mode, "RGB")

        # Act / Assert
        self.assertRaises(ValueError, im.convert, mode="CMYK", matrix=matrix)

    def test_matrix_wrong_mode(self):
        # Arrange
        im = hopper("L")
        # fmt: off
        matrix = (
            0.412453, 0.357580, 0.180423, 0,
            0.212671, 0.715160, 0.072169, 0,
            0.019334, 0.119193, 0.950227, 0)
        # fmt: on
        self.assertEqual(im.mode, "L")

        # Act / Assert
        self.assertRaises(ValueError, im.convert, mode="L", matrix=matrix)

    def test_matrix_xyz(self):
        def matrix_convert(mode):
            # Arrange
            im = hopper("RGB")
            im.info["transparency"] = (255, 0, 0)
            # fmt: off
            matrix = (
                0.412453, 0.357580, 0.180423, 0,
                0.212671, 0.715160, 0.072169, 0,
                0.019334, 0.119193, 0.950227, 0)
            # fmt: on
            self.assertEqual(im.mode, "RGB")

            # Act
            # Convert an RGB image to the CIE XYZ colour space
            converted_im = im.convert(mode=mode, matrix=matrix)

            # Assert
            self.assertEqual(converted_im.mode, mode)
            self.assertEqual(converted_im.size, im.size)
            with Image.open("Tests/images/hopper-XYZ.png") as target:
                if converted_im.mode == "RGB":
                    self.assert_image_similar(converted_im, target, 3)
                    self.assertEqual(converted_im.info["transparency"], (105, 54, 4))
                else:
                    self.assert_image_similar(converted_im, target.getchannel(0), 1)
                    self.assertEqual(converted_im.info["transparency"], 105)

        matrix_convert("RGB")
        matrix_convert("L")

    def test_matrix_identity(self):
        # Arrange
        im = hopper("RGB")
        # fmt: off
        identity_matrix = (
            1, 0, 0, 0,
            0, 1, 0, 0,
            0, 0, 1, 0)
        # fmt: on
        self.assertEqual(im.mode, "RGB")

        # Act
        # Convert with an identity matrix
        converted_im = im.convert(mode="RGB", matrix=identity_matrix)

        # Assert
        # No change
        self.assert_image_equal(converted_im, im)



from PIL import Image, ImageSequence, TiffImagePlugin

from .helper import PillowTestCase, hopper


class TestImageSequence(PillowTestCase):
    def test_sanity(self):

        test_file = self.tempfile("temp.im")

        im = hopper("RGB")
        im.save(test_file)

        seq = ImageSequence.Iterator(im)

        index = 0
        for frame in seq:
            self.assert_image_equal(im, frame)
            self.assertEqual(im.tell(), index)
            index += 1

        self.assertEqual(index, 1)

        self.assertRaises(AttributeError, ImageSequence.Iterator, 0)

    def test_iterator(self):
        with Image.open("Tests/images/multipage.tiff") as im:
            i = ImageSequence.Iterator(im)
            for index in range(0, im.n_frames):
                self.assertEqual(i[index], next(i))
            self.assertRaises(IndexError, lambda: i[index + 1])
            self.assertRaises(StopIteration, next, i)

    def test_iterator_min_frame(self):
        with Image.open("Tests/images/hopper.psd") as im:
            i = ImageSequence.Iterator(im)
            for index in range(1, im.n_frames):
                self.assertEqual(i[index], next(i))

    def _test_multipage_tiff(self):
        with Image.open("Tests/images/multipage.tiff") as im:
            for index, frame in enumerate(ImageSequence.Iterator(im)):
                frame.load()
                self.assertEqual(index, im.tell())
                frame.convert("RGB")

    def test_tiff(self):
        self._test_multipage_tiff()

    def test_libtiff(self):
        codecs = dir(Image.core)

        if "libtiff_encoder" not in codecs or "libtiff_decoder" not in codecs:
            self.skipTest("tiff support not available")

        TiffImagePlugin.READ_LIBTIFF = True
        self._test_multipage_tiff()
        TiffImagePlugin.READ_LIBTIFF = False

    def test_consecutive(self):
        with Image.open("Tests/images/multipage.tiff") as im:
            firstFrame = None
            for frame in ImageSequence.Iterator(im):
                if firstFrame is None:
                    firstFrame = frame.copy()
            for frame in ImageSequence.Iterator(im):
                self.assert_image_equal(frame, firstFrame)
                break

    def test_palette_mmap(self):
        # Using mmap in ImageFile can require to reload the palette.
        with Image.open("Tests/images/multipage-mmap.tiff") as im:
            color1 = im.getpalette()[0:3]
            im.seek(0)
            color2 = im.getpalette()[0:3]
            self.assertEqual(color1, color2)

    def test_all_frames(self):
        # Test a single image
        with Image.open("Tests/images/iss634.gif") as im:
            ims = ImageSequence.all_frames(im)

            self.assertEqual(len(ims), 42)
            for i, im_frame in enumerate(ims):
                self.assertFalse(im_frame is im)

                im.seek(i)
                self.assert_image_equal(im, im_frame)

            # Test a series of images
            ims = ImageSequence.all_frames([im, hopper(), im])
            self.assertEqual(len(ims), 85)

            # Test an operation
            ims = ImageSequence.all_frames(im, lambda im_frame: im_frame.rotate(90))
            for i, im_frame in enumerate(ims):
                im.seek(i)
                self.assert_image_equal(im.rotate(90), im_frame)



from PIL import Image

from .helper import PillowTestCase, hopper


class TestImageGetExtrema(PillowTestCase):
    def test_extrema(self):
        def extrema(mode):
            return hopper(mode).getextrema()

        self.assertEqual(extrema("1"), (0, 255))
        self.assertEqual(extrema("L"), (1, 255))
        self.assertEqual(extrema("I"), (1, 255))
        self.assertEqual(extrema("F"), (1, 255))
        self.assertEqual(extrema("P"), (0, 225))  # fixed palette
        self.assertEqual(extrema("RGB"), ((0, 255), (0, 255), (0, 255)))
        self.assertEqual(extrema("RGBA"), ((0, 255), (0, 255), (0, 255), (255, 255)))
        self.assertEqual(extrema("CMYK"), ((0, 255), (0, 255), (0, 255), (0, 0)))
        self.assertEqual(extrema("I;16"), (1, 255))

    def test_true_16(self):
        with Image.open("Tests/images/16_bit_noise.tif") as im:
            self.assertEqual(im.mode, "I;16")
            extrema = im.getextrema()
        self.assertEqual(extrema, (106, 285))



import unittest

from PIL import Image, ImagePalette, features

from .helper import PillowTestCase, hopper

try:
    from PIL import MicImagePlugin
except ImportError:
    olefile_installed = False
else:
    olefile_installed = True

TEST_FILE = "Tests/images/hopper.mic"


@unittest.skipUnless(olefile_installed, "olefile package not installed")
@unittest.skipUnless(features.check("libtiff"), "libtiff not installed")
class TestFileMic(PillowTestCase):
    def test_sanity(self):
        with Image.open(TEST_FILE) as im:
            im.load()
            self.assertEqual(im.mode, "RGBA")
            self.assertEqual(im.size, (128, 128))
            self.assertEqual(im.format, "MIC")

            # Adjust for the gamma of 2.2 encoded into the file
            lut = ImagePalette.make_gamma_lut(1 / 2.2)
            im = Image.merge("RGBA", [chan.point(lut) for chan in im.split()])

            im2 = hopper("RGBA")
            self.assert_image_similar(im, im2, 10)

    def test_n_frames(self):
        with Image.open(TEST_FILE) as im:

            self.assertEqual(im.n_frames, 1)

    def test_is_animated(self):
        with Image.open(TEST_FILE) as im:

            self.assertFalse(im.is_animated)

    def test_tell(self):
        with Image.open(TEST_FILE) as im:

            self.assertEqual(im.tell(), 0)

    def test_seek(self):
        with Image.open(TEST_FILE) as im:

            im.seek(0)
            self.assertEqual(im.tell(), 0)

            self.assertRaises(EOFError, im.seek, 99)
            self.assertEqual(im.tell(), 0)

    def test_invalid_file(self):
        # Test an invalid OLE file
        invalid_file = "Tests/images/flower.jpg"
        self.assertRaises(SyntaxError, MicImagePlugin.MicImageFile, invalid_file)

        # Test a valid OLE file, but not a MIC file
        ole_file = "Tests/images/test-ole-file.doc"
        self.assertRaises(SyntaxError, MicImagePlugin.MicImageFile, ole_file)



import base64
import io
import itertools
import logging
import os
from collections import namedtuple
from ctypes import c_float

from PIL import Image, TiffImagePlugin, TiffTags, features

from .helper import PillowTestCase, hopper

logger = logging.getLogger(__name__)


class LibTiffTestCase(PillowTestCase):
    def setUp(self):
        if not features.check("libtiff"):
            self.skipTest("tiff support not available")

    def _assert_noerr(self, im):
        """Helper tests that assert basic sanity about the g4 tiff reading"""
        # 1 bit
        self.assertEqual(im.mode, "1")

        # Does the data actually load
        im.load()
        im.getdata()

        try:
            self.assertEqual(im._compression, "group4")
        except AttributeError:
            print("No _compression")
            print(dir(im))

        # can we write it back out, in a different form.
        out = self.tempfile("temp.png")
        im.save(out)

        out_bytes = io.BytesIO()
        im.save(out_bytes, format="tiff", compression="group4")


class TestFileLibTiff(LibTiffTestCase):
    def test_g4_tiff(self):
        """Test the ordinary file path load path"""

        test_file = "Tests/images/hopper_g4_500.tif"
        with Image.open(test_file) as im:
            self.assertEqual(im.size, (500, 500))
            self._assert_noerr(im)

    def test_g4_large(self):
        test_file = "Tests/images/pport_g4.tif"
        with Image.open(test_file) as im:
            self._assert_noerr(im)

    def test_g4_tiff_file(self):
        """Testing the string load path"""

        test_file = "Tests/images/hopper_g4_500.tif"
        with open(test_file, "rb") as f:
            with Image.open(f) as im:
                self.assertEqual(im.size, (500, 500))
                self._assert_noerr(im)

    def test_g4_tiff_bytesio(self):
        """Testing the stringio loading code path"""
        test_file = "Tests/images/hopper_g4_500.tif"
        s = io.BytesIO()
        with open(test_file, "rb") as f:
            s.write(f.read())
            s.seek(0)
        with Image.open(s) as im:
            self.assertEqual(im.size, (500, 500))
            self._assert_noerr(im)

    def test_g4_non_disk_file_object(self):
        """Testing loading from non-disk non-BytesIO file object"""
        test_file = "Tests/images/hopper_g4_500.tif"
        s = io.BytesIO()
        with open(test_file, "rb") as f:
            s.write(f.read())
            s.seek(0)
        r = io.BufferedReader(s)
        with Image.open(r) as im:
            self.assertEqual(im.size, (500, 500))
            self._assert_noerr(im)

    def test_g4_eq_png(self):
        """ Checking that we're actually getting the data that we expect"""
        with Image.open("Tests/images/hopper_bw_500.png") as png:
            with Image.open("Tests/images/hopper_g4_500.tif") as g4:
                self.assert_image_equal(g4, png)

    # see https://github.com/python-pillow/Pillow/issues/279
    def test_g4_fillorder_eq_png(self):
        """ Checking that we're actually getting the data that we expect"""
        with Image.open("Tests/images/g4-fillorder-test.png") as png:
            with Image.open("Tests/images/g4-fillorder-test.tif") as g4:
                self.assert_image_equal(g4, png)

    def test_g4_write(self):
        """Checking to see that the saved image is the same as what we wrote"""
        test_file = "Tests/images/hopper_g4_500.tif"
        with Image.open(test_file) as orig:
            out = self.tempfile("temp.tif")
            rot = orig.transpose(Image.ROTATE_90)
            self.assertEqual(rot.size, (500, 500))
            rot.save(out)

            with Image.open(out) as reread:
                self.assertEqual(reread.size, (500, 500))
                self._assert_noerr(reread)
                self.assert_image_equal(reread, rot)
                self.assertEqual(reread.info["compression"], "group4")

                self.assertEqual(reread.info["compression"], orig.info["compression"])

                self.assertNotEqual(orig.tobytes(), reread.tobytes())

    def test_adobe_deflate_tiff(self):
        test_file = "Tests/images/tiff_adobe_deflate.tif"
        with Image.open(test_file) as im:
            self.assertEqual(im.mode, "RGB")
            self.assertEqual(im.size, (278, 374))
            self.assertEqual(im.tile[0][:3], ("libtiff", (0, 0, 278, 374), 0))
            im.load()

            self.assert_image_equal_tofile(im, "Tests/images/tiff_adobe_deflate.png")

    def test_write_metadata(self):
        """ Test metadata writing through libtiff """
        for legacy_api in [False, True]:
            f = self.tempfile("temp.tiff")
            with Image.open("Tests/images/hopper_g4.tif") as img:
                img.save(f, tiffinfo=img.tag)

                if legacy_api:
                    original = img.tag.named()
                else:
                    original = img.tag_v2.named()

            # PhotometricInterpretation is set from SAVE_INFO,
            # not the original image.
            ignored = [
                "StripByteCounts",
                "RowsPerStrip",
                "PageNumber",
                "PhotometricInterpretation",
            ]

            with Image.open(f) as loaded:
                if legacy_api:
                    reloaded = loaded.tag.named()
                else:
                    reloaded = loaded.tag_v2.named()

            for tag, value in itertools.chain(reloaded.items(), original.items()):
                if tag not in ignored:
                    val = original[tag]
                    if tag.endswith("Resolution"):
                        if legacy_api:
                            self.assertEqual(
                                c_float(val[0][0] / val[0][1]).value,
                                c_float(value[0][0] / value[0][1]).value,
                                msg="%s didn't roundtrip" % tag,
                            )
                        else:
                            self.assertEqual(
                                c_float(val).value,
                                c_float(value).value,
                                msg="%s didn't roundtrip" % tag,
                            )
                    else:
                        self.assertEqual(val, value, msg="%s didn't roundtrip" % tag)

            # https://github.com/python-pillow/Pillow/issues/1561
            requested_fields = ["StripByteCounts", "RowsPerStrip", "StripOffsets"]
            for field in requested_fields:
                self.assertIn(field, reloaded, "%s not in metadata" % field)

    def test_additional_metadata(self):
        # these should not crash. Seriously dummy data, most of it doesn't make
        # any sense, so we're running up against limits where we're asking
        # libtiff to do stupid things.

        # Get the list of the ones that we should be able to write

        core_items = {
            tag: info
            for tag, info in ((s, TiffTags.lookup(s)) for s in TiffTags.LIBTIFF_CORE)
            if info.type is not None
        }

        # Exclude ones that have special meaning
        # that we're already testing them
        with Image.open("Tests/images/hopper_g4.tif") as im:
            for tag in im.tag_v2:
                try:
                    del core_items[tag]
                except KeyError:
                    pass

            # Type codes:
            #     2: "ascii",
            #     3: "short",
            #     4: "long",
            #     5: "rational",
            #     12: "double",
            # Type: dummy value
            values = {
                2: "test",
                3: 1,
                4: 2 ** 20,
                5: TiffImagePlugin.IFDRational(100, 1),
                12: 1.05,
            }

            new_ifd = TiffImagePlugin.ImageFileDirectory_v2()
            for tag, info in core_items.items():
                if info.length == 1:
                    new_ifd[tag] = values[info.type]
                if info.length == 0:
                    new_ifd[tag] = tuple(values[info.type] for _ in range(3))
                else:
                    new_ifd[tag] = tuple(values[info.type] for _ in range(info.length))

            # Extra samples really doesn't make sense in this application.
            del new_ifd[338]

            out = self.tempfile("temp.tif")
            TiffImagePlugin.WRITE_LIBTIFF = True

            im.save(out, tiffinfo=new_ifd)

        TiffImagePlugin.WRITE_LIBTIFF = False

    def test_custom_metadata(self):
        tc = namedtuple("test_case", "value,type,supported_by_default")
        custom = {
            37000 + k: v
            for k, v in enumerate(
                [
                    tc(4, TiffTags.SHORT, True),
                    tc(123456789, TiffTags.LONG, True),
                    tc(-4, TiffTags.SIGNED_BYTE, False),
                    tc(-4, TiffTags.SIGNED_SHORT, False),
                    tc(-123456789, TiffTags.SIGNED_LONG, False),
                    tc(TiffImagePlugin.IFDRational(4, 7), TiffTags.RATIONAL, True),
                    tc(4.25, TiffTags.FLOAT, True),
                    tc(4.25, TiffTags.DOUBLE, True),
                    tc("custom tag value", TiffTags.ASCII, True),
                    tc(b"custom tag value", TiffTags.BYTE, True),
                    tc((4, 5, 6), TiffTags.SHORT, True),
                    tc((123456789, 9, 34, 234, 219387, 92432323), TiffTags.LONG, True),
                    tc((-4, 9, 10), TiffTags.SIGNED_BYTE, False),
                    tc((-4, 5, 6), TiffTags.SIGNED_SHORT, False),
                    tc(
                        (-123456789, 9, 34, 234, 219387, -92432323),
                        TiffTags.SIGNED_LONG,
                        False,
                    ),
                    tc((4.25, 5.25), TiffTags.FLOAT, True),
                    tc((4.25, 5.25), TiffTags.DOUBLE, True),
                    # array of TIFF_BYTE requires bytes instead of tuple for backwards
                    # compatibility
                    tc(bytes([4]), TiffTags.BYTE, True),
                    tc(bytes((4, 9, 10)), TiffTags.BYTE, True),
                ]
            )
        }

        libtiffs = [False]
        if Image.core.libtiff_support_custom_tags:
            libtiffs.append(True)

        for libtiff in libtiffs:
            TiffImagePlugin.WRITE_LIBTIFF = libtiff

            def check_tags(tiffinfo):
                im = hopper()

                out = self.tempfile("temp.tif")
                im.save(out, tiffinfo=tiffinfo)

                with Image.open(out) as reloaded:
                    for tag, value in tiffinfo.items():
                        reloaded_value = reloaded.tag_v2[tag]
                        if (
                            isinstance(reloaded_value, TiffImagePlugin.IFDRational)
                            and libtiff
                        ):
                            # libtiff does not support real RATIONALS
                            self.assertAlmostEqual(float(reloaded_value), float(value))
                            continue

                        if libtiff and isinstance(value, bytes):
                            value = value.decode()

                        self.assertEqual(reloaded_value, value)

            # Test with types
            ifd = TiffImagePlugin.ImageFileDirectory_v2()
            for tag, tagdata in custom.items():
                ifd[tag] = tagdata.value
                ifd.tagtype[tag] = tagdata.type
            check_tags(ifd)

            # Test without types. This only works for some types, int for example are
            # always encoded as LONG and not SIGNED_LONG.
            check_tags(
                {
                    tag: tagdata.value
                    for tag, tagdata in custom.items()
                    if tagdata.supported_by_default
                }
            )
        TiffImagePlugin.WRITE_LIBTIFF = False

    def test_int_dpi(self):
        # issue #1765
        im = hopper("RGB")
        out = self.tempfile("temp.tif")
        TiffImagePlugin.WRITE_LIBTIFF = True
        im.save(out, dpi=(72, 72))
        TiffImagePlugin.WRITE_LIBTIFF = False
        with Image.open(out) as reloaded:
            self.assertEqual(reloaded.info["dpi"], (72.0, 72.0))

    def test_g3_compression(self):
        with Image.open("Tests/images/hopper_g4_500.tif") as i:
            out = self.tempfile("temp.tif")
            i.save(out, compression="group3")

            with Image.open(out) as reread:
                self.assertEqual(reread.info["compression"], "group3")
                self.assert_image_equal(reread, i)

    def test_little_endian(self):
        with Image.open("Tests/images/16bit.deflate.tif") as im:
            self.assertEqual(im.getpixel((0, 0)), 480)
            self.assertEqual(im.mode, "I;16")

            b = im.tobytes()
            # Bytes are in image native order (little endian)
            self.assertEqual(b[0], ord(b"\xe0"))
            self.assertEqual(b[1], ord(b"\x01"))

            out = self.tempfile("temp.tif")
            # out = "temp.le.tif"
            im.save(out)
        with Image.open(out) as reread:
            self.assertEqual(reread.info["compression"], im.info["compression"])
            self.assertEqual(reread.getpixel((0, 0)), 480)
        # UNDONE - libtiff defaults to writing in native endian, so
        # on big endian, we'll get back mode = 'I;16B' here.

    def test_big_endian(self):
        with Image.open("Tests/images/16bit.MM.deflate.tif") as im:
            self.assertEqual(im.getpixel((0, 0)), 480)
            self.assertEqual(im.mode, "I;16B")

            b = im.tobytes()

            # Bytes are in image native order (big endian)
            self.assertEqual(b[0], ord(b"\x01"))
            self.assertEqual(b[1], ord(b"\xe0"))

            out = self.tempfile("temp.tif")
            im.save(out)
            with Image.open(out) as reread:
                self.assertEqual(reread.info["compression"], im.info["compression"])
                self.assertEqual(reread.getpixel((0, 0)), 480)

    def test_g4_string_info(self):
        """Tests String data in info directory"""
        test_file = "Tests/images/hopper_g4_500.tif"
        with Image.open(test_file) as orig:
            out = self.tempfile("temp.tif")

            orig.tag[269] = "temp.tif"
            orig.save(out)

        with Image.open(out) as reread:
            self.assertEqual("temp.tif", reread.tag_v2[269])
            self.assertEqual("temp.tif", reread.tag[269][0])

    def test_12bit_rawmode(self):
        """ Are we generating the same interpretation
        of the image as Imagemagick is? """
        TiffImagePlugin.READ_LIBTIFF = True
        with Image.open("Tests/images/12bit.cropped.tif") as im:
            im.load()
            TiffImagePlugin.READ_LIBTIFF = False
            # to make the target --
            # convert 12bit.cropped.tif -depth 16 tmp.tif
            # convert tmp.tif -evaluate RightShift 4 12in16bit2.tif
            # imagemagick will auto scale so that a 12bit FFF is 16bit FFF0,
            # so we need to unshift so that the integer values are the same.

            self.assert_image_equal_tofile(im, "Tests/images/12in16bit.tif")

    def test_blur(self):
        # test case from irc, how to do blur on b/w image
        # and save to compressed tif.
        from PIL import ImageFilter

        out = self.tempfile("temp.tif")
        with Image.open("Tests/images/pport_g4.tif") as im:
            im = im.convert("L")

        im = im.filter(ImageFilter.GaussianBlur(4))
        im.save(out, compression="tiff_adobe_deflate")

        with Image.open(out) as im2:
            im2.load()

            self.assert_image_equal(im, im2)

    def test_compressions(self):
        # Test various tiff compressions and assert similar image content but reduced
        # file sizes.
        im = hopper("RGB")
        out = self.tempfile("temp.tif")
        im.save(out)
        size_raw = os.path.getsize(out)

        for compression in ("packbits", "tiff_lzw"):
            im.save(out, compression=compression)
            size_compressed = os.path.getsize(out)
            with Image.open(out) as im2:
                self.assert_image_equal(im, im2)

        im.save(out, compression="jpeg")
        size_jpeg = os.path.getsize(out)
        with Image.open(out) as im2:
            self.assert_image_similar(im, im2, 30)

        im.save(out, compression="jpeg", quality=30)
        size_jpeg_30 = os.path.getsize(out)
        with Image.open(out) as im3:
            self.assert_image_similar(im2, im3, 30)

        self.assertGreater(size_raw, size_compressed)
        self.assertGreater(size_compressed, size_jpeg)
        self.assertGreater(size_jpeg, size_jpeg_30)

    def test_quality(self):
        im = hopper("RGB")
        out = self.tempfile("temp.tif")

        self.assertRaises(ValueError, im.save, out, compression="tiff_lzw", quality=50)
        self.assertRaises(ValueError, im.save, out, compression="jpeg", quality=-1)
        self.assertRaises(ValueError, im.save, out, compression="jpeg", quality=101)
        self.assertRaises(ValueError, im.save, out, compression="jpeg", quality="good")
        im.save(out, compression="jpeg", quality=0)
        im.save(out, compression="jpeg", quality=100)

    def test_cmyk_save(self):
        im = hopper("CMYK")
        out = self.tempfile("temp.tif")

        im.save(out, compression="tiff_adobe_deflate")
        with Image.open(out) as im2:
            self.assert_image_equal(im, im2)

    def xtest_bw_compression_w_rgb(self):
        """ This test passes, but when running all tests causes a failure due
            to output on stderr from the error thrown by libtiff. We need to
            capture that but not now"""

        im = hopper("RGB")
        out = self.tempfile("temp.tif")

        self.assertRaises(IOError, im.save, out, compression="tiff_ccitt")
        self.assertRaises(IOError, im.save, out, compression="group3")
        self.assertRaises(IOError, im.save, out, compression="group4")

    def test_fp_leak(self):
        im = Image.open("Tests/images/hopper_g4_500.tif")
        fn = im.fp.fileno()

        os.fstat(fn)
        im.load()  # this should close it.
        self.assertRaises(OSError, os.fstat, fn)
        im = None  # this should force even more closed.
        self.assertRaises(OSError, os.fstat, fn)
        self.assertRaises(OSError, os.close, fn)

    def test_multipage(self):
        # issue #862
        TiffImagePlugin.READ_LIBTIFF = True
        with Image.open("Tests/images/multipage.tiff") as im:
            # file is a multipage tiff,  10x10 green, 10x10 red, 20x20 blue

            im.seek(0)
            self.assertEqual(im.size, (10, 10))
            self.assertEqual(im.convert("RGB").getpixel((0, 0)), (0, 128, 0))
            self.assertTrue(im.tag.next)

            im.seek(1)
            self.assertEqual(im.size, (10, 10))
            self.assertEqual(im.convert("RGB").getpixel((0, 0)), (255, 0, 0))
            self.assertTrue(im.tag.next)

            im.seek(2)
            self.assertFalse(im.tag.next)
            self.assertEqual(im.size, (20, 20))
            self.assertEqual(im.convert("RGB").getpixel((0, 0)), (0, 0, 255))

        TiffImagePlugin.READ_LIBTIFF = False

    def test_multipage_nframes(self):
        # issue #862
        TiffImagePlugin.READ_LIBTIFF = True
        with Image.open("Tests/images/multipage.tiff") as im:
            frames = im.n_frames
            self.assertEqual(frames, 3)
            for _ in range(frames):
                im.seek(0)
                # Should not raise ValueError: I/O operation on closed file
                im.load()

        TiffImagePlugin.READ_LIBTIFF = False

    def test__next(self):
        TiffImagePlugin.READ_LIBTIFF = True
        with Image.open("Tests/images/hopper.tif") as im:
            self.assertFalse(im.tag.next)
            im.load()
            self.assertFalse(im.tag.next)

    def test_4bit(self):
        # Arrange
        test_file = "Tests/images/hopper_gray_4bpp.tif"
        original = hopper("L")

        # Act
        TiffImagePlugin.READ_LIBTIFF = True
        with Image.open(test_file) as im:
            TiffImagePlugin.READ_LIBTIFF = False

            # Assert
            self.assertEqual(im.size, (128, 128))
            self.assertEqual(im.mode, "L")
            self.assert_image_similar(im, original, 7.3)

    def test_gray_semibyte_per_pixel(self):
        test_files = (
            (
                24.8,  # epsilon
                (  # group
                    "Tests/images/tiff_gray_2_4_bpp/hopper2.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper2I.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper2R.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper2IR.tif",
                ),
            ),
            (
                7.3,  # epsilon
                (  # group
                    "Tests/images/tiff_gray_2_4_bpp/hopper4.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper4I.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper4R.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper4IR.tif",
                ),
            ),
        )
        original = hopper("L")
        for epsilon, group in test_files:
            with Image.open(group[0]) as im:
                self.assertEqual(im.size, (128, 128))
                self.assertEqual(im.mode, "L")
                self.assert_image_similar(im, original, epsilon)
            for file in group[1:]:
                with Image.open(file) as im2:
                    self.assertEqual(im2.size, (128, 128))
                    self.assertEqual(im2.mode, "L")
                    self.assert_image_equal(im, im2)

    def test_save_bytesio(self):
        # PR 1011
        # Test TIFF saving to io.BytesIO() object.

        TiffImagePlugin.WRITE_LIBTIFF = True
        TiffImagePlugin.READ_LIBTIFF = True

        # Generate test image
        pilim = hopper()

        def save_bytesio(compression=None):

            buffer_io = io.BytesIO()
            pilim.save(buffer_io, format="tiff", compression=compression)
            buffer_io.seek(0)

            with Image.open(buffer_io) as pilim_load:
                self.assert_image_similar(pilim, pilim_load, 0)

        save_bytesio()
        save_bytesio("raw")
        save_bytesio("packbits")
        save_bytesio("tiff_lzw")

        TiffImagePlugin.WRITE_LIBTIFF = False
        TiffImagePlugin.READ_LIBTIFF = False

    def test_crashing_metadata(self):
        # issue 1597
        with Image.open("Tests/images/rdf.tif") as im:
            out = self.tempfile("temp.tif")

            TiffImagePlugin.WRITE_LIBTIFF = True
            # this shouldn't crash
            im.save(out, format="TIFF")
        TiffImagePlugin.WRITE_LIBTIFF = False

    def test_page_number_x_0(self):
        # Issue 973
        # Test TIFF with tag 297 (Page Number) having value of 0 0.
        # The first number is the current page number.
        # The second is the total number of pages, zero means not available.
        outfile = self.tempfile("temp.tif")
        # Created by printing a page in Chrome to PDF, then:
        # /usr/bin/gs -q -sDEVICE=tiffg3 -sOutputFile=total-pages-zero.tif
        # -dNOPAUSE /tmp/test.pdf -c quit
        infile = "Tests/images/total-pages-zero.tif"
        with Image.open(infile) as im:
            # Should not divide by zero
            im.save(outfile)

    def test_fd_duplication(self):
        # https://github.com/python-pillow/Pillow/issues/1651

        tmpfile = self.tempfile("temp.tif")
        with open(tmpfile, "wb") as f:
            with open("Tests/images/g4-multi.tiff", "rb") as src:
                f.write(src.read())

        im = Image.open(tmpfile)
        im.n_frames
        im.close()
        # Should not raise PermissionError.
        os.remove(tmpfile)

    def test_read_icc(self):
        with Image.open("Tests/images/hopper.iccprofile.tif") as img:
            icc = img.info.get("icc_profile")
            self.assertIsNotNone(icc)
        TiffImagePlugin.READ_LIBTIFF = True
        with Image.open("Tests/images/hopper.iccprofile.tif") as img:
            icc_libtiff = img.info.get("icc_profile")
            self.assertIsNotNone(icc_libtiff)
        TiffImagePlugin.READ_LIBTIFF = False
        self.assertEqual(icc, icc_libtiff)

    def test_multipage_compression(self):
        with Image.open("Tests/images/compression.tif") as im:

            im.seek(0)
            self.assertEqual(im._compression, "tiff_ccitt")
            self.assertEqual(im.size, (10, 10))

            im.seek(1)
            self.assertEqual(im._compression, "packbits")
            self.assertEqual(im.size, (10, 10))
            im.load()

            im.seek(0)
            self.assertEqual(im._compression, "tiff_ccitt")
            self.assertEqual(im.size, (10, 10))
            im.load()

    def test_save_tiff_with_jpegtables(self):
        # Arrange
        outfile = self.tempfile("temp.tif")

        # Created with ImageMagick: convert hopper.jpg hopper_jpg.tif
        # Contains JPEGTables (347) tag
        infile = "Tests/images/hopper_jpg.tif"
        with Image.open(infile) as im:
            # Act / Assert
            # Should not raise UnicodeDecodeError or anything else
            im.save(outfile)

    def test_16bit_RGB_tiff(self):
        with Image.open("Tests/images/tiff_16bit_RGB.tiff") as im:
            self.assertEqual(im.mode, "RGB")
            self.assertEqual(im.size, (100, 40))
            self.assertEqual(
                im.tile,
                [
                    (
                        "libtiff",
                        (0, 0, 100, 40),
                        0,
                        ("RGB;16N", "tiff_adobe_deflate", False, 8),
                    )
                ],
            )
            im.load()

            self.assert_image_equal_tofile(im, "Tests/images/tiff_16bit_RGB_target.png")

    def test_16bit_RGBa_tiff(self):
        with Image.open("Tests/images/tiff_16bit_RGBa.tiff") as im:
            self.assertEqual(im.mode, "RGBA")
            self.assertEqual(im.size, (100, 40))
            self.assertEqual(
                im.tile,
                [
                    (
                        "libtiff",
                        (0, 0, 100, 40),
                        0,
                        ("RGBa;16N", "tiff_lzw", False, 38236),
                    )
                ],
            )
            im.load()

            self.assert_image_equal_tofile(
                im, "Tests/images/tiff_16bit_RGBa_target.png"
            )

    def test_gimp_tiff(self):
        # Read TIFF JPEG images from GIMP [@PIL168]

        codecs = dir(Image.core)
        if "jpeg_decoder" not in codecs:
            self.skipTest("jpeg support not available")

        filename = "Tests/images/pil168.tif"
        with Image.open(filename) as im:
            self.assertEqual(im.mode, "RGB")
            self.assertEqual(im.size, (256, 256))
            self.assertEqual(
                im.tile,
                [("libtiff", (0, 0, 256, 256), 0, ("RGB", "jpeg", False, 5122))],
            )
            im.load()

            self.assert_image_equal_tofile(im, "Tests/images/pil168.png")

    def test_sampleformat(self):
        # https://github.com/python-pillow/Pillow/issues/1466
        with Image.open("Tests/images/copyleft.tiff") as im:
            self.assertEqual(im.mode, "RGB")

            self.assert_image_equal_tofile(im, "Tests/images/copyleft.png", mode="RGB")

    def test_lzw(self):
        with Image.open("Tests/images/hopper_lzw.tif") as im:
            self.assertEqual(im.mode, "RGB")
            self.assertEqual(im.size, (128, 128))
            self.assertEqual(im.format, "TIFF")
            im2 = hopper()
            self.assert_image_similar(im, im2, 5)

    def test_strip_cmyk_jpeg(self):
        infile = "Tests/images/tiff_strip_cmyk_jpeg.tif"
        with Image.open(infile) as im:
            self.assert_image_similar_tofile(
                im, "Tests/images/pil_sample_cmyk.jpg", 0.5
            )

    def test_strip_cmyk_16l_jpeg(self):
        infile = "Tests/images/tiff_strip_cmyk_16l_jpeg.tif"
        with Image.open(infile) as im:
            self.assert_image_similar_tofile(
                im, "Tests/images/pil_sample_cmyk.jpg", 0.5
            )

    def test_strip_ycbcr_jpeg_2x2_sampling(self):
        infile = "Tests/images/tiff_strip_ycbcr_jpeg_2x2_sampling.tif"
        with Image.open(infile) as im:
            self.assert_image_similar_tofile(im, "Tests/images/flower.jpg", 0.5)

    def test_strip_ycbcr_jpeg_1x1_sampling(self):
        infile = "Tests/images/tiff_strip_ycbcr_jpeg_1x1_sampling.tif"
        with Image.open(infile) as im:
            self.assert_image_equal_tofile(im, "Tests/images/flower2.jpg")

    def test_tiled_cmyk_jpeg(self):
        infile = "Tests/images/tiff_tiled_cmyk_jpeg.tif"
        with Image.open(infile) as im:
            self.assert_image_similar_tofile(
                im, "Tests/images/pil_sample_cmyk.jpg", 0.5
            )

    def test_tiled_ycbcr_jpeg_1x1_sampling(self):
        infile = "Tests/images/tiff_tiled_ycbcr_jpeg_1x1_sampling.tif"
        with Image.open(infile) as im:
            self.assert_image_equal_tofile(im, "Tests/images/flower2.jpg")

    def test_tiled_ycbcr_jpeg_2x2_sampling(self):
        infile = "Tests/images/tiff_tiled_ycbcr_jpeg_2x2_sampling.tif"
        with Image.open(infile) as im:
            self.assert_image_similar_tofile(im, "Tests/images/flower.jpg", 0.5)

    def test_old_style_jpeg(self):
        infile = "Tests/images/old-style-jpeg-compression.tif"
        with Image.open(infile) as im:
            self.assert_image_equal_tofile(
                im, "Tests/images/old-style-jpeg-compression.png"
            )

    def test_no_rows_per_strip(self):
        # This image does not have a RowsPerStrip TIFF tag
        infile = "Tests/images/no_rows_per_strip.tif"
        with Image.open(infile) as im:
            im.load()
        self.assertEqual(im.size, (950, 975))

    def test_orientation(self):
        with Image.open("Tests/images/g4_orientation_1.tif") as base_im:
            for i in range(2, 9):
                with Image.open("Tests/images/g4_orientation_" + str(i) + ".tif") as im:
                    im.load()

                    self.assert_image_similar(base_im, im, 0.7)

    def test_sampleformat_not_corrupted(self):
        # Assert that a TIFF image with SampleFormat=UINT tag is not corrupted
        # when saving to a new file.
        # Pillow 6.0 fails with "OSError: cannot identify image file".
        tiff = io.BytesIO(
            base64.b64decode(
                b"SUkqAAgAAAAPAP4ABAABAAAAAAAAAAABBAABAAAAAQAAAAEBBAABAAAAAQAA"
                b"AAIBAwADAAAAwgAAAAMBAwABAAAACAAAAAYBAwABAAAAAgAAABEBBAABAAAA"
                b"4AAAABUBAwABAAAAAwAAABYBBAABAAAAAQAAABcBBAABAAAACwAAABoBBQAB"
                b"AAAAyAAAABsBBQABAAAA0AAAABwBAwABAAAAAQAAACgBAwABAAAAAQAAAFMB"
                b"AwADAAAA2AAAAAAAAAAIAAgACAABAAAAAQAAAAEAAAABAAAAAQABAAEAAAB4"
                b"nGNgYAAAAAMAAQ=="
            )
        )
        out = io.BytesIO()
        with Image.open(tiff) as im:
            im.save(out, format="tiff")
        out.seek(0)
        with Image.open(out) as im:
            im.load()

    def test_realloc_overflow(self):
        TiffImagePlugin.READ_LIBTIFF = True
        with Image.open("Tests/images/tiff_overflow_rows_per_strip.tif") as im:
            with self.assertRaises(IOError) as e:
                im.load()

            # Assert that the error code is IMAGING_CODEC_MEMORY
            self.assertEqual(str(e.exception), "-9")
        TiffImagePlugin.READ_LIBTIFF = False



import unittest
from io import BytesIO

from PIL import Image

from .helper import PillowTestCase, is_win32

# Limits for testing the leak
mem_limit = 1024 * 1048576
stack_size = 8 * 1048576
iterations = int((mem_limit / stack_size) * 2)
codecs = dir(Image.core)
test_file = "Tests/images/rgb_trns_ycbc.jp2"


@unittest.skipIf(is_win32(), "requires Unix or macOS")
class TestJpegLeaks(PillowTestCase):
    def setUp(self):
        if "jpeg2k_encoder" not in codecs or "jpeg2k_decoder" not in codecs:
            self.skipTest("JPEG 2000 support not available")

    def test_leak_load(self):
        from resource import setrlimit, RLIMIT_AS, RLIMIT_STACK

        setrlimit(RLIMIT_STACK, (stack_size, stack_size))
        setrlimit(RLIMIT_AS, (mem_limit, mem_limit))
        for _ in range(iterations):
            with Image.open(test_file) as im:
                im.load()

    def test_leak_save(self):
        from resource import setrlimit, RLIMIT_AS, RLIMIT_STACK

        setrlimit(RLIMIT_STACK, (stack_size, stack_size))
        setrlimit(RLIMIT_AS, (mem_limit, mem_limit))
        for _ in range(iterations):
            with Image.open(test_file) as im:
                im.load()
                test_output = BytesIO()
                im.save(test_output, "JPEG2000")
                test_output.seek(0)
                test_output.read()


if __name__ == "__main__":
    unittest.main()



from PIL import Image

from .helper import PillowTestCase, hopper

# sample ppm stream
test_file = "Tests/images/hopper.ppm"


class TestFilePpm(PillowTestCase):
    def test_sanity(self):
        with Image.open(test_file) as im:
            im.load()
            self.assertEqual(im.mode, "RGB")
            self.assertEqual(im.size, (128, 128))
            self.assertEqual(im.format, "PPM")
            self.assertEqual(im.get_format_mimetype(), "image/x-portable-pixmap")

    def test_16bit_pgm(self):
        with Image.open("Tests/images/16_bit_binary.pgm") as im:
            im.load()
            self.assertEqual(im.mode, "I")
            self.assertEqual(im.size, (20, 100))
            self.assertEqual(im.get_format_mimetype(), "image/x-portable-graymap")

            with Image.open("Tests/images/16_bit_binary_pgm.png") as tgt:
                self.assert_image_equal(im, tgt)

    def test_16bit_pgm_write(self):
        with Image.open("Tests/images/16_bit_binary.pgm") as im:
            im.load()

            f = self.tempfile("temp.pgm")
            im.save(f, "PPM")

            with Image.open(f) as reloaded:
                self.assert_image_equal(im, reloaded)

    def test_pnm(self):
        with Image.open("Tests/images/hopper.pnm") as im:
            self.assert_image_similar(im, hopper(), 0.0001)

            f = self.tempfile("temp.pnm")
            im.save(f)

            with Image.open(f) as reloaded:
                self.assert_image_equal(im, reloaded)

    def test_truncated_file(self):
        path = self.tempfile("temp.pgm")
        with open(path, "w") as f:
            f.write("P6")

        self.assertRaises(ValueError, Image.open, path)

    def test_neg_ppm(self):
        # Storage.c accepted negative values for xsize, ysize.  the
        # internal open_ppm function didn't check for sanity but it
        # has been removed. The default opener doesn't accept negative
        # sizes.

        with self.assertRaises(IOError):
            Image.open("Tests/images/negative_size.ppm")

    def test_mimetypes(self):
        path = self.tempfile("temp.pgm")

        with open(path, "w") as f:
            f.write("P4\n128 128\n255")
        with Image.open(path) as im:
            self.assertEqual(im.get_format_mimetype(), "image/x-portable-bitmap")

        with open(path, "w") as f:
            f.write("PyCMYK\n128 128\n255")
        with Image.open(path) as im:
            self.assertEqual(im.get_format_mimetype(), "image/x-portable-anymap")
